#include "asset.h"
#include "memory.h"
#include "plat.h"
#include "profile.h"
#include "render.h"
#include "standard.h"

static BM_Font* main_font;
static Fixed_Font smol_font;

static const Asset_Meta asset_meta[] = {
#define x(id, type, path) { type, path },
	assets_xmacro()
#undef x
};

Asset* loaded_assets[asset_id_count];

typedef struct {
	Colour pixel;
	unsigned short offset, size;
} Run;

static void init_bitmap_asset(Asset* asset, char* raw, int size) {
	short w, h;
	int us, rc, rcm1, i, j, dst, src, e;
	Bitmap* bmp;
	Colour* pixels;
	const Run* runs, * run, * nxt;
	profiler_begin("decompress_bitmap");
	w = *(short*)(raw + 0);
	h = *(short*)(raw + 2);
	rc = *(int*)(raw + 4);
	us = w * h;
	bmp = stack_alloc(sizeof *bmp);
	bmp->pixels = stack_alloc(us * 4);
	bmp->w = w;
	bmp->h = h;
	runs = (const Run*)(raw + 8);
	pixels = (Colour*)(raw + 8 + rc * sizeof(Run));
	rcm1 = rc - 1;
	dst = src = 0;
	run = &runs[0];
	for (i = 0; i < run->offset; i++)
		bmp->pixels[dst++] = pixels[src++];
	for (i = 0; i < rcm1; i++) {
		run = &runs[i];
		nxt = &runs[i + 1];
		for (j = 0; j < run->size; j++)
			bmp->pixels[dst++] = run->pixel;
		e = nxt->offset - (run->offset + run->size);
		for (j = 0; j < e; j++)
			bmp->pixels[dst++] = pixels[src++];
	}
	run = &runs[i];
	for (i = 0; i < run->size; i++)
		bmp->pixels[dst++] = run->pixel;
	asset->payload = bmp;
	asset->payload_size = size;
	profiler_end();
}

static void init_sound_asset(Asset* asset, char* raw, int size) {
	asset->payload = raw;
	asset->payload_size = size;
}

static void init_txt_asset(Asset* asset, void* raw, int size) {
	Text* text;
	text = stack_alloc(sizeof *text);
	text->text = raw;
	text->len = size;
	asset->payload = text;
	asset->payload_size = size;
}

/* Loading assets isn't that efficient. Too bad. */
static int get_asset_info(
	File* pack,
	int pack_count,
	const char* name,
	int* offset,
	int* size
) {
	int head_size = 8, i;
	char buf[56];
	file_seek(pack, head_size, file_whence_begin);
	for (i = 0; i < pack_count; i++) {
		file_read(pack, size,   4);
		file_read(pack, offset, 4);
		file_read(pack, buf,    sizeof buf);
		if (string_equal(buf, name)) {
			return 1;
		}
	}
	return 0;
}

Asset_Data* read_asset_data(
	const char* name,
	Arena* arena
) {
	Asset_Data* r = 0;
	File* pack;
	int pack_count, offset, size;
	profiler_begin("read_asset_data");
	pack = file_open("pack", file_flags_read);
	if (!pack) {
		print_err("Couldn't find 'pack'. Did you delete it?\n");
		pbreak(error_file_not_found);
		profiler_end();
		return 0;
	}
	file_read(pack, &pack_count, 4);
	if (get_asset_info(pack, pack_count, name, &offset, &size)) {
		r = arena_alloc(arena, sizeof *r);
		r->offset = offset;
		r->size = size;
		r->buffer = arena_alloc(arena, r->size);
		file_seek(pack, offset, file_whence_begin);
		file_read(pack, r->buffer, r->size);
	} else {
		print_err(\
			"%s doesn't exist in the pack. "
			"Is the pack corrupt?\n",
			name
		);
		file_close(pack);
		profiler_end();
		return 0;
	}
	file_close(pack);
	profiler_end();
	return r;
}

static BM_Font* load_font(
	Bitmap* bmp,
	BM_Font_Range range
) {
	int s;
	BM_Font* r;
	s = bm_font_size(range);
	r = stack_alloc(s);
	init_bm_font(r, bmp, range);
	return r;
}

void preload_assets() {
	int offset, size, s, pack_count, id;
	Asset* asset;
	File* pack;
	const Asset_Meta* meta;
	const char* name;
	Asset_Type type;
	void* raw;
	profiler_begin("preload_assets");
	pack = file_open("pack", file_flags_read);
	if (!pack) {
		print_err("Couldn't find 'pack'. Did you delete it?\n");
		pbreak(error_file_not_found);
	}
	file_read(pack, &pack_count, 4);
#define gi(e_) \
		meta = get_asset_meta((Asset_ID)id); \
		name = meta->name; \
		type = meta->type; \
		if (get_asset_info( \
			pack, \
			pack_count, \
			name, \
			&offset, \
			&size \
		)) { \
			file_seek(pack, offset, file_whence_begin); \
			e_ \
		} else { \
			print_err( \
				"%s doesn't exist in the pack. " \
				"Is the pack corrupt?\n", \
				name \
			); \
			pbreak(error_file_not_found); \
		}
	for (id = 0; id < asset_id_count; id++) {
		gi({
			asset = (Asset*)stack_alloc(sizeof *asset);
			asset->id = (Asset_ID)id;
			asset->type = type;
			s = type == asset_type_txt ? size + 1 : size;
			raw = galloc(s);
			file_seek(pack, offset, file_whence_begin);
			file_read(pack, raw, size);
			switch (type) {
				case asset_type_bitmap:
					init_bitmap_asset(asset, raw, size);
					gfree(raw);
					break;
				case asset_type_txt:
					((char*)raw)[size] = 0;
					init_txt_asset(asset, raw, size);
					break;
				case asset_type_sound:
					init_sound_asset(asset, raw, size);
					break;
				default:
					print_err("Invalid asset type.\n");
					pbreak(error_invalid_asset);
					break;
			}
			loaded_assets[id] = asset;
		});
	}
#undef gi
	file_close(pack);
	main_font = load_font(
		(Bitmap*)get_bitmap(asset_id_main_font),
		bm_font_range_printable
	);
	init_fixed_font(
		&smol_font,
		get_bitmap(asset_id_smol_font),
		4,
		4,
		1
	);
	profiler_end();
}

const Asset* get_asset(Asset_ID id) {
	return loaded_assets[id];
}

const Asset_Meta* get_asset_meta(Asset_ID id) {
	return &asset_meta[id];
}

const Bitmap* get_bitmap(Asset_ID id) {
	const Bitmap* b;
	const Asset* asset;

	asset = get_asset(id);
	b = asset->payload;
	return b;
}

const Text* get_text(Asset_ID id) {
	const Text* t;
	const Asset* asset;

	asset = get_asset(id);
	t = asset->payload;
	return t;
}

const struct BM_Font* get_main_font() {
	return main_font;
}

const struct Fixed_Font* get_smol_font() {
	return &smol_font;
}
