#include "arms.h"
#include "asset.h"
#include "boss.h"
#include "fx.h"
#include "gamestate.h"
#include "maths.h"
#include "rcache.h"
#include "sprite.h"
#include "world.h"

static Entity new_cia_body(
	World* world,
	int x,
	int y
) {
	Entity e;
	int id;
	CSprite* spr;
	CPosition* pos;
	e = new_entity(world);
	id = entity_index(e);
	add_components(
		world,
		e,
		ctype_position |
		ctype_sprite
	);
	pos = &world->positions[id];
	spr = &world->sprites[id];
	pos->x = x - (30 << fbits);
	pos->y = y + (55 << fbits);
	init_csprite(spr, sprite_cia_body);
	return e;
}

static Entity new_boss_cia(
	World* world,
	int x,
	int y
) {
	Entity e;
	int id;
	CSprite* spr;
	CPosition* pos;
	CCollider* col;
	CMonster* mons;
	CBody* body;
	CBoss* boss;
	CVelocity* vel;
	CDamageable* dam;
	CList* list;
	e = new_entity(world);
	id = entity_index(e);
	add_components(
		world,
		e,
		ctype_position   |
		ctype_sprite     |
		ctype_monster    |
		ctype_body       |
		ctype_collider   |
		ctype_velocity   |
		ctype_list       |
		ctype_damageable |
		ctype_boss
	);
	add_tags(
		world,
		e,
		tag_enemy |
		tag_bio
	);
	pos = &world->positions[id];
	spr = &world->sprites[id];
	mons = &world->monsters[id];
	body = &world->bodies[id];
	col = &world->colliders[id];
	vel = &world->velocities[id];
	list = &world->lists[id];
	boss = &world->bosses[id];
	dam = &world->damageables[id];
	boss->type = boss_type_cia;
	dam->hp = 100;
	dam->hurting = 0;
	vel->x = 0;
	vel->y = 0;
	col->x = 10 << fbits;
	col->y = 6  << fbits;
	col->w = 39 << fbits;
	col->h = 57 << fbits;
	pos->x = x;
	pos->y = y;
	body->grounded = 0;
	mons->timer = 0;
	mons->state = 0;
	list->next = new_cia_body(world, x, y);
	init_csprite(spr, sprite_cia_head);
	return e;
}

Entity new_boss(
	Boss_Type type,
	struct World* world,
	int x,
	int y
) {
	switch (type) {
		case boss_type_cia:
			return new_boss_cia(world, x, y);
		default: return 0;
	}
}

void draw_healthbar(int hp) {
	const Sprite* l, * r, * f, * m;
	const Bitmap* b;
	const int px = 12, py = 6;
	int i, y, d, a;
	int s = px;
	int e = renderer_w - px;
	l = get_sprite(sprite_boss_hb_left);
	r = get_sprite(sprite_boss_hb_right);
	f = get_sprite(sprite_boss_hb_middle_full);
	m = get_sprite(sprite_boss_hb_middle);
	/* Assume they're all on the same bitmap. */
	b = get_bitmap(l->bitmap);
	/* ..and that they're all the same height. */
	y = renderer_h - py - l->rect.h;
	rc_add_cmd_bmp(
		b,
		s,
		y,
		&l->rect
	);
	d = (e - r->rect.w) - (s + l->rect.w);
	a = ((hp << 16) / 100) * d >> 16;
	for (i = 0; i < a; i++)
		rc_add_cmd_bmp(
			b,
			s + l->rect.w + i * m->rect.w,
			y,
			&f->rect
		);
	for (; i < d; i++)
		rc_add_cmd_bmp(
			b,
			s + l->rect.w + i * m->rect.w,
			y,
			&m->rect
		);
	rc_add_cmd_bmp(
		b,
		e - r->rect.w,
		y,
		&r->rect
	);
}

static void cia_shoot(
	World* world,
	int x,
	int y,
	int dx,
	int dy
) {
	Entity e;
	int id;
	CPosition* pos;
	CSprite* sprite;
	CVelocity* velo;
	CCollider* col;
	CBody* body;
	CImpact* imp;
	if (!dx && !dy) return;
	e = new_entity(world);
	id = entity_index(e);
	add_components(
		world,
		e,
		ctype_position |
		ctype_velocity |
		ctype_sprite |
		ctype_collider |
		ctype_body |
		ctype_impact
	);
	add_tags(
		world,
		e,
		tag_enemy |
		tag_destroy_offscreen |
		tag_bullet |
		tag_destroy_collision
	);
	pos = &world->positions[id];
	sprite = &world->sprites[id];
	velo = &world->velocities[id];
	col = &world->colliders[id];
	body = &world->bodies[id];
	imp = &world->impacts[id];
	pos->x = x;
	pos->y = y;
	velo->x = dx;
	velo->y = dy;
	vec_scale(&velo->x, &velo->y, cia_bullet_speed);
	col->x = 0;
	col->y = 0;
	col->w = 3 << fbits;
	col->h = 3 << fbits;
	body->grounded = 0;
	imp->anim = animation_cia_bullet_hit;
	init_csprite(sprite, sprite_cia_bullet);
	sprite->z = 100;
}

static void update_cia(
	Gamestate* gamestate,
	World* world,
	Entity e
) {
	int id, pid, dx, dy, x, y, t, rx, ry, st, ct, i;
	int bid;
	CSprite* spr;
	CPosition* pos, * bpos;
	CPosition* ppos;
	CDamageable* dam;
	CMonster* mons;
	CList* list;
	Player* player;
	Entity body;
	id = entity_index(e);
	pos = &world->positions[id];
	spr = &world->sprites[id];
	dam = &world->damageables[id];
	player = &world->player;
	pid = entity_index(player->e);
	ppos = &world->positions[pid];
	mons = &world->monsters[id];
	list = &world->lists[id];
	if (dam->hp <= 0) {
		dam->hp = 0;
		remove_components(
			world,
			e,
			ctype_damageable
		);
		if (mons->timer % 2)
			init_csprite(spr, sprite_cia_head_hurt);
		else
			init_csprite(spr, sprite_cia_head);
		body = list->next;
		bid = entity_index(body);
		bpos = &world->positions[bid];
		pos->y += f1;
		bpos->y += f1;
		if (pos->y >> fbits > renderer_h) {
			transition_level(
				gamestate,
				"",
				world,
				gamestate_type_credit
			);
		}
		return;
	}
	if (mons->timer % 50 == 0) {
		int os = mons->state;
		while (mons->state == os)
			mons->state = (rng()) % 5;
	}
	x = pos->x + (32 << fbits);
	y = pos->y + (36 << fbits);
	if (
		mons->state <= 3 &&
		mons->timer % 5 == 0 &&
		!player->dead
	) {
		init_csprite(spr, sprite_cia_head_fire);
		dx = ppos->x - x;
		dy = ppos->y - y;
		vec_minise(&dx, &dy);
		vec_nrmise(&dx, &dy);
		if (mons->state == 0) {
			cia_shoot(world, x, y,  dx,  dy);
		} else if (mons->state == 1) {
			cia_shoot(world, x, y,  dx,  dy);
			cia_shoot(world, x, y, -dx, -dy);
			cia_shoot(world, x, y, -dx,  dy);
			cia_shoot(world, x, y,  dx, -dy);
		} else if (mons->state == 2) {
			cia_shoot(world, x, y,  dx,  dy);
			cia_shoot(world, x, y, -dx, -dy);
		} else if (mons->state == 3) {
			t = mons->timer % f1;
			for (i = 0; i < 5; i++) {
				st = fsin(t + i * 100);
				ct = fcos(t + i * 100);
				dx = f1;
				dy = 0;
				rx =
					((dx * ct) >> fbits) -
					((dy * st) >> fbits);
				ry =
					((dx * st) >> fbits) +
					((dy * ct) >> fbits);
				cia_shoot(world, x, y,  rx,  ry);
				cia_shoot(world, x, y, -rx, -ry);
			}
		}
	} else if (
			mons->state == 4 &&
			mons->timer % 30 == 0
	) {
		dx = cia_bomb_vel_x;
		dy = cia_bomb_vel_y;
		new_bomb(world, x + (35 << fbits), y, dx, dy);
		new_bomb(world, x - (30 << fbits), y, -dx, dy);
	} else if (dam->hurting % 2)
		init_csprite(spr, sprite_cia_head_hurt);
	else 
		init_csprite(spr, sprite_cia_head);
	if (dam->hurting == 1) {
		new_blood_splatter(
			world,
			pos->x + (spr->rect.w << fbits) / 2,
			pos->y + (spr->rect.h << fbits) / 2,
			0,
			0
		);
	}
}

static void update_boss(
	Gamestate* gamestate,
	World* world,
	Entity e,
	const CBoss* boss
) {
	switch (boss->type) {
		case boss_type_cia:
			update_cia(gamestate, world, e);
			break;
		default: break;
	}
}

void boss_system(Gamestate* gamestate, World* world) {
	general_system_vars
	const CBoss* boss;
	Entity e;
	general_system(ctype_boss)
		boss = &world->bosses[i];
		e = world->entities[i];
		update_boss(gamestate, world, e, boss);
	}
}

void boss_render_system(World* world) {
	general_system_vars
	const CDamageable* dam;
	general_system(ctype_boss | ctype_damageable)
		dam = &world->damageables[i];
		draw_healthbar(dam->hp);
	}
}
