#include <stdio.h>
#include <stdarg.h>

#include "config.h"

#include "animation.c"
#include "boss_type.c"
#include "cfgparse.c"
#include "mapobject.c"
#include "memory.c"
#include "sprite.c"
#include "std_printers.c"

#define obj_arena_size level_loader_memory
#define max_map_objects 256

Arena obj_arena;
Map_Object* map_objects[max_map_objects];
int map_object_count = 0;

typedef struct {
	unsigned char tile;
	unsigned short offset, size;
} Run;

static void process_level_meta(
	const cfg_Object* obj,
	FILE* outfile
) {
	int w = 32, h = 32, spx = 0, spy = 0, lc = 4;
	char name[28];
	const cfg_Prop* prop;
	for (prop = obj->props; prop; prop = prop->next) {
		if (prop->type == cfg_type_int) {
			if (!strcmp("width", prop->name)) {
				w = prop->as.num;
			} else if (!strcmp("height", prop->name)) {
				h = prop->as.num;
			} else if (!strcmp("spawn_x", prop->name)) {
				spx = prop->as.num;
			} else if (!strcmp("spawn_y", prop->name)) {
				spy = prop->as.num;
			}
		} else if (prop->type == cfg_type_string) {
			if (!strcmp(prop->name, "name")) {
				strcpy(name, prop->as.str);
			}
		}
	}
	fwrite(name, sizeof name, 1, outfile);
	fwrite(&lc,  4, 1, outfile);
	fwrite(&w,   4, 1, outfile);
	fwrite(&h,   4, 1, outfile);
	fwrite(&spx, 4, 1, outfile);
	fwrite(&spy, 4, 1, outfile);
}

static void re_encode(
	unsigned char* src,
	int s,
	FILE* outfile
) {
	unsigned char* buf;
	Run* runs, run;
	int j, i, rc = 0, bs = 0, t;
	runs = malloc(s * sizeof *runs);
	buf = malloc(s);
	for (i = 0; i < s; i++) {
		t = src[i];
		if (i + 1 < s) {
			run.tile = t;
			run.offset = i;
			run.size = 1;
			for (
				j = i + 1;
				j < s && t == src[j];
				j++, run.size++
			);
			if (run.size >= sizeof(Run)) {
				i += run.size - 1;
				runs[rc++] = run;
			} else
				buf[bs++] = t;
		} else
			buf[bs++] = t;
	}
	fwrite(
		&rc,
		sizeof rc,
		1,
		outfile
	);
	fwrite(
		&bs,
		sizeof bs,
		1,
		outfile
	);
	fwrite(
		runs,
		sizeof *runs,
		rc,
		outfile
	);
	fwrite(
		buf,
		1,
		bs,
		outfile
	);
	free(runs);
	free(buf);
}

static void process_level_layer(
	const cfg_Object* obj,
	FILE* outfile
) {
	const cfg_Prop* pindex;
	const cfg_Prop* pdata;
	pindex = find_prop_of(obj, "index", cfg_type_int);
	pdata  = find_prop_of(obj, "data",  cfg_type_data);
	fwrite(&pindex->as.num, 4, 1, outfile);
	re_encode(
		(unsigned char*)pdata->as.data.data,
		pdata->as.data.size,
		outfile
	);
}

static void process_level_colliders(
	const cfg_Object* obj,
	FILE* outfile
) {
	const int n1 = -1;
	const cfg_Prop* pdata;
	pdata  = find_prop_of(obj, "data",  cfg_type_data);
	fwrite(&n1, 4, 1, outfile);
	re_encode(
		(unsigned char*)pdata->as.data.data,
		pdata->as.data.size,
		outfile
	);
}

static void process_level_object(const cfg_Object* obj) {
	Map_Object* mo;
	const char* name;
	name = find_prop_of(
		obj,
		"type",
		cfg_type_string
	)->as.str;
	mo = create_map_object(
		name,
		obj,
		&obj_arena
	);
	map_objects[map_object_count++] = mo;
}

static void write_level_objects(FILE* outfile) {
	int i, size;
	fwrite(&map_object_count, 4, 1, outfile);
	for (i = 0; i < map_object_count; i++) {
		size = get_map_object_size(map_objects[i]->type);
		fwrite(map_objects[i], size, 1, outfile);
	}
}

int main(int argc, char** argv) {
	int size;
	FILE* infile, * outfile;
	char* src, * obj_buf;
	cfg_Object* obj;
	Arena arena;

	if (argc < 3) {
		fprintf(stderr, "Usage: %s infile outfile.\n", argv[0]);
		return 1;
	}

	infile = fopen(argv[1], "r");
	if (!infile) {
		fprintf(stderr, "Failed to open %s.\n", argv[1]);
		return 2;
	}

	outfile = fopen(argv[2], "wb");
	if (!outfile) {
		fprintf(stderr, "Failed to open %s.\n", argv[2]);
		return 3;
	}

	fseek(infile, 0, SEEK_END);
	size = ftell(infile);
	rewind(infile);

	src = malloc(size);
	size = fread(src, 1, size, infile);
	src[size] = 0;

	init_arena(
		&arena,
		malloc(level_loader_memory),
		level_loader_memory);
	obj = cfg_parse(src, &arena);

	fwrite("LEVL", 4, 1, outfile);

	obj_buf = malloc(obj_arena_size);
	init_arena(&obj_arena, obj_buf, obj_arena_size);
	while (obj) {
		if (!strcmp("level", obj->name)) {
			process_level_meta(obj, outfile);
		} else if (!strcmp("layer", obj->name)) {
			process_level_layer(obj, outfile);
		} else if (!strcmp("colliders", obj->name)) {
			process_level_colliders(obj, outfile);
		} else if (!strcmp("object", obj->name)) {
			process_level_object(obj);
		}
		obj = obj->next;
	}
	write_level_objects(outfile);

	free(src);
	free(arena.buffer);
	free(obj_arena.buffer);

	fclose(outfile);
	fclose(infile);
	return 0;
}
