#include <string.h>

#include "asset.h"
#include "cutscene.h"
#include "error.h"
#include "plat.h"
#include "rcache.h"
#include "world.h"

static Com_Bitset get_cmd_mask(Cut_Cmd_Type t) {
	switch (t) {
#define x(name, com) \
	case cut_cmd_ ## name: return com;
		cut_cmd_type_xmacro();
	default: return 0;
	}
#undef x
}

const char* get_cut_cmd_type_name(
	Cut_Cmd_Type t
) {
	switch (t) {
#define x(name, com) \
	case cut_cmd_ ## name: return #name;
		cut_cmd_type_xmacro()
#undef x
		default: return 0;
	}
}

Cut_Cmd_Type get_cut_cmd_type_from_name(
	const char* name
) {
#define x(n, com) \
	if (!strcmp(name, #n)) \
		return cut_cmd_ ## n;
	cut_cmd_type_xmacro()
#undef x
	return cut_cmd_type_count;
}

void init_cutscene(Cutscene* s) {
	s->buf = 0;
	s->buf_size = 0;
	s->endframe = 0;
}

void load_cutscene(
	Cutscene* s,
	const char* name,
	Arena* arena
) {
	Asset_Data* d;
	int ptr;
	d = read_asset_data(
		name,
		arena
	);
	if (memcmp(d->buffer, "CUTS", 4)) {
		print_err("Invalid cutscene file.\n");
		pbreak(error_invalid_asset);
	}
	ptr = 4;
	s->endframe = *(int*)(d->buffer + ptr);
	ptr += 4;
	s->buf = d->buffer + ptr;
	s->buf_size = d->size - 8;
}

void play_cutscene(Cutscene* s) {
	s->frame = 0;
	s->cmd = (Cut_Cmd*)s->buf;
	s->running = 0;
	s->paused = 0;
}

static void finish_cmd(
	Cutscene* s,
	int idx
) {
	if (s->running > 1)
		s->run[idx] = s->run[s->running];
	s->running--;
}

static void render_dialog(
	Cutscene* s,
	int idx,
	Cut_Cmd_Dialog* cmd
) {
	const Sprite* spr;
	const BM_Font* f;
	int l;
	char buf[64];
	Rectangle r = {
		20,
		20,
		0,
		0
	};
	Rectangle b = {
		10,
		10,
		renderer_w - 20,
		84
	};
	r.w = r.x + 60;
	r.h = r.y + 60;
	f = get_main_font();
	rc_add_cmd_rect(
		&b,
		make_black()
	);
	if (cmd->face == sprite_face_null) {
		b.x -= 74;
		goto skip_face;
	}
	spr = get_sprite(cmd->face);
	rc_add_cmd_clip(&r);
	rc_add_cmd_bmp(
		get_bitmap(spr->bitmap),
		r.x + 3,
		r.y + 3,
		&spr->rect
	);
	rc_add_cmd_reset_clip();
	spr = get_sprite(sprite_face_border);
	rc_add_cmd_bmp(
		get_bitmap(spr->bitmap),
		r.x,
		r.y,
		&spr->rect
	);
skip_face:
	strcpy(buf, (const char*)(cmd + 1));
	l = strlen(buf);
	buf[mini(s->typewriter, l)] = 0;
	rc_add_cmd_rfont_text(
		f,
		b.x + 84,
		b.y + b.h / 2 - f->height / 2,
		buf
	);
	if (btn_just_released(btn_jump)) {
		s->paused = 0;
		finish_cmd(s, idx);
	}
	s->typewriter++;
}

void render_cutscene(Cutscene* s) {
	int i;
	Cut_Cmd* cmd;
	for (i = 0; i < s->running; i++) {
		cmd = s->run[i];
		switch (cmd->type) {
		case cut_cmd_dialog:
			render_dialog(s, i, (Cut_Cmd_Dialog*)cmd);
			break;
		default:
			break;
		}
	}
}

void play_cmd(Cutscene* s, Cut_Cmd* cmd) {
	if (cmd->type == cut_cmd_dialog)
		s->paused = 1;
#if DEBUG
	if (s->running >= cutscene_max_running) {
		print_err("Too many concurrent cutscene commands.");
		pbreak(error_out_of_memory);
	}
#endif
	s->run[s->running++] = cmd;
}

#define for_with_handle(m, h) \
	general_system(m | ctype_cutscene) \
		if (strcmp(world->cutscenes[i].handle, h)) \
			continue;

static void exec_moveto(
	Cutscene* s,
	int idx,
	Cut_Cmd_Moveto* c,
	World* world
) {
	general_system_vars
	Com_Bitset m;
	int dx, dy, l;
	CPosition* pos;
	(void)s;
	m = get_cmd_mask(c->cmd.type);
	for_with_handle(m, c->handle)
		pos = &world->positions[i];
		dx = c->x - pos->x;
		dy = c->y - pos->y;
		l = vec_len(dx, dy);
		if (
			(!dx && !dy) ||
			(l > 0 && l < c->speed)
		) {
			finish_cmd(s, idx);
		} else if (dx || dy) {
			vec_minise(&dx, &dy);
			vec_nrmise(&dx, &dy);
			vec_scale(&dx, &dy, c->speed);
			pos->x += dx;
			pos->y += dy;
		}
	}
}

static void exec_cmd(
	Cutscene* s,
	int idx,
	Cut_Cmd* c,
	World* world
) {
	s->typewriter = 0;
	switch (c->type) {
		case cut_cmd_moveto:
			exec_moveto(s, idx, (Cut_Cmd_Moveto*)c, world);
			break;
		default: break;
	}
}

void update_cutscene(Cutscene* s, World* world) {
	Cut_Cmd* cmd, * end;
	int i;
	cmd = s->cmd;
	end = (Cut_Cmd*)(s->buf + s->buf_size);
	if (!s->paused) {
		while (cmd != end) {
			if (cmd->frame == s->frame)
				play_cmd(s, cmd);
			cmd = (Cut_Cmd*)((char*)cmd + cmd->size);
		}
		for (i = 0; i < s->running; i++) {
			exec_cmd(s, i, s->run[i], world);
		}
		s->frame++;
	}
}

int cutscene_done(Cutscene* s) {
	return s->frame >= s->endframe;
}

