#include "asset.h"
#include "boss.h"
#include "cfgparse.h"
#include "cutparse.h"
#include "cutscene.h"
#include "editor.h"
#include "item.h"
#include "levelloader.h"
#include "mapobject.h"
#include "memory.h"
#include "microui.h"
#include "plat.h"
#include "rcache.h"
#include "world.h"

#include <stdio.h>
#include <string.h>

#if enable_editor && enable_imgui

static mu_Style editor_style = {
	/* font | size | padding | spacing | indent */
	NULL, { 84, 0 }, 3, 3, 24,
	/* title_height | scrollbar_size | thumb_size */
	6, 2, 8,
	{
		{ 0,   0,   0,   255 }, /* MU_COLOR_TEXT */
		{ 0,   0,   0,   255 }, /* MU_COLOR_BORDER */
		{ 255, 255, 255, 255 }, /* MU_COLOR_WINDOWBG */
		{ 240, 248, 255, 255 }, /* MU_COLOR_TITLEBG */
		{ 0,   0,   0,   255 }, /* MU_COLOR_TITLETEXT */
		{ 240, 248, 255, 0   }, /* MU_COLOR_PANELBG */
		{ 240, 248, 255, 255 }, /* MU_COLOR_BUTTON */
		{ 240, 248, 255, 255 }, /* MU_COLOR_BUTTONHOVER */
		{ 200, 200, 255, 255 }, /* MU_COLOR_BUTTONFOCUS */
		{ 240, 248, 255, 255 }, /* MU_COLOR_BASE */
		{ 240, 248, 255, 255 }, /* MU_COLOR_BASEHOVER */
		{ 200, 200, 255, 255 }, /* MU_COLOR_BASEFOCUS */
		{ 240, 248, 255, 255 }, /* MU_COLOR_SCROLLBASE */
		{ 200, 200, 255, 255 }  /* MU_COLOR_SCROLLTHUMB */
	}
};

static const Map_Object_Editor mo_editors[] = {
#define x(type, factory, editor, spawner, params) \
	editor,
map_objects_xmacro()
#undef x
};

typedef void (*Map_Object_Saver)(
	const Map_Object*,
	FILE*
);
#define x(type, factory, editor, spawner, params) \
	void type ## _saver( \
		const Map_Object* obj, \
		FILE* file \
	);
map_objects_xmacro()
#undef x

static const Map_Object_Saver mo_savers[] = {
#define x(type, factory, editor, spawner, params) \
	type ## _saver,
map_objects_xmacro()
#undef x
};

#define mo_arena_size 1024
#define ct_arena_size (1024 * 1024)
#define max_mapobject_instances 256

extern int pause_gameplay;

#define editor_tool_xmacro \
	x(editor_tool_paint,    "Paint") \
	x(editor_tool_erase,    "Erase") \
	x(editor_tool_collider, "Collider") \
	x(editor_tool_picker,   "Pick")

static const char* layer_names[] = {
#define x(t, n) n,
	map_layers_xmacro
#undef x
};

typedef enum {
#define x(e, n) \
	e,
	editor_tool_xmacro
#undef x
	editor_tool_count
} Editor_Tool;

mu_Context* mu;

typedef struct {
	Map_Object* obj;
	Entity e;
} Map_Object_Instance;

typedef struct {
	int spx, spy;
} Level_Meta;

struct {
	int tile, show, showcols, showspawn, mo_count;
	int show_level, show_cutscene;
	int doffx, doffy;
	int ctcmd_frame;
	int colmask;
	Editor_Tool tool;
	Map_Layer_ID layer;
	Level_Meta lm;
	Arena mo_arena;
	Arena ct_arena;
	Cutscene ct;
	Cut_Cmd* ctcmd;
	Map_Object_Instance mos[max_mapobject_instances];
	Map_Object_Instance* selected_mo, * dragging_mo;
	char name[32];
	char cname[32];
	char lname[28];
} ed;

static int text_width(
	mu_Font font,
	const char* text,
	int len
) {
	const char* c;
	(void)text;
	(void)font;
	if (len == -1) {
		len = 0;
		for (c = text; *c; c++, len++);
	}
	return (get_smol_font()->char_w + 1) * len;
}

static int text_height(mu_Font font) {
	(void)font;
	return get_smol_font()->char_h;
}

static void save_map_objects(FILE* file) {
	int i;
	const Map_Object* obj;
	for (i = 0; i < ed.mo_count; i++) {
		obj = ed.mos[i].obj;
		fprintf(file, "[object]\n");
		fprintf(
			file,
			"type: \"%s\n",
			get_map_object_name(obj->type)
		);
		mo_savers[obj->type](obj, file);
	}
}

static int ed_save_level(const World* world, const char* name) {
	FILE* file;
	const Map* map;
	const Map_Layer* layer;
	int i, j, s;
	const char* c;
	file = fopen(name, "w");
	if (!file) return 0;
	map = &world->map;
	fprintf(
		file,
		"[level]\nname: \"%s\nwidth: %d\nheight: %d\nlayer_count: %d\n"
		"spawn_x: %d\nspawn_y: %d\n",
		ed.lname,
		map->w,
		map->h,
		map_layer_count,
		ed.lm.spx,
		ed.lm.spy
	);
	save_map_objects(file);
	s = map->w * map->h * sizeof *map->layers[0].tiles;
	for (i = 0; i < map_layer_count; i++) {
		layer = &map->layers[i];
		fprintf(file, "[layer]\n");
		fprintf(file, "index: %d\n", i);
		fprintf(file, "data: $");
		for (
			j = 0, c = (const char*)layer->tiles;
			j < s;
			j++, c++
		) {
			fprintf(file, "%02x", *c);
		}
		fprintf(file, "\n");
	}
	fprintf(file, "[colliders]\n");
	fprintf(file, "data: $");
	s = map->w * map->h * sizeof *map->colliders;
	for (
		i = 0, c = (const char*)map->colliders;
		i < s;
		i++, c++
	) {
		fprintf(file, "%02x", *c);
	}
	fprintf(file, "\n");
	fclose(file);
	return 1;
}

static void process_level_meta(
	World* world,
	const cfg_Object* obj
) {
	int w = 32, h = 32;
	const cfg_Prop* prop;
	for (prop = obj->props; prop; prop = prop->next) {
		if (prop->type == cfg_type_int) {
			if (!strcmp("width", prop->name)) {
				w = prop->as.num;
			} else if (!strcmp("height", prop->name)) {
				h = prop->as.num;
			} else if (!strcmp("spawn_x", prop->name)) {
				ed.lm.spx = prop->as.num;
			} else if (!strcmp("spawn_y", prop->name)) {
				ed.lm.spy = prop->as.num;
			}
		} else if (prop->type == cfg_type_string) {
			if (!strcmp(prop->name, "name")) {
				strcpy(ed.lname, prop->as.str);
			}
		}
	}
	preload_init(world, w, h);
}

static void process_level_layer(
	World* world,
	const cfg_Object* obj
) {
	const cfg_Prop* pindex;
	const cfg_Prop* pdata;
	Map_Layer* layer;
	pindex = find_prop_of(obj, "index", cfg_type_int);
	pdata  = find_prop_of(obj, "data",  cfg_type_data);
	layer = &world->map.layers[pindex->as.num];
	memcpy(
		layer->tiles,
		pdata->as.data.data,
		pdata->as.data.size
	);
}

static void process_level_colliders(
	World* world,
	const cfg_Object* obj
) {
	const cfg_Prop* pdata;
	pdata = find_prop_of(obj, "data", cfg_type_data);
	memcpy(
		&world->map.colliders,
		pdata->as.data.data,
		pdata->as.data.size
	);
}

static void instance_map_object(
	World* world,
	Map_Object* obj
) {
	Entity e;
	Map_Object_Instance* i;
	e = spawn_mapobject(world, obj);
#ifdef DEBUG
	if (ed.mo_count >= max_mapobject_instances) {
		print_err("Too many map object instances.\n");
		pbreak(error_out_of_memory);
	}
#endif
	i = &ed.mos[ed.mo_count++];
	i->e = e;
	i->obj = obj;
}

static void process_level_object(
	World* world,
	const cfg_Object* obj
) {
	Map_Object* mo;
	const char* name;
	name = find_prop_of(
		obj,
		"type",
		cfg_type_string
	)->as.str;
	mo = create_map_object(name, obj, &ed.mo_arena);
	instance_map_object(world, mo);
}

void destroy_mo(World* world, Map_Object_Instance* mo) {
	destroy_entity(world, mo->e);
	*mo = ed.mos[ed.mo_count - 1];
	ed.mo_count--;
}

void dup_mo(World* world, Map_Object_Instance* mo) {
	int size;
	Map_Object* new;
	size = get_map_object_size(mo->obj->type);
	new = arena_alloc(&ed.mo_arena, size);
	memcpy(new, mo->obj, size);
	instance_map_object(world, new);
}

void init_new_map_object(Map_Object* mo) {
	prop_paramstruct*      prparams;
	killzone_paramstruct*  kzparams;
	conttrig_paramstruct*  ctparams;
	gamesaver_paramstruct* gsparams;
	pickup_paramstruct*    piparams;
	cuttrig_paramstruct*   cuparams;
	boss_paramstruct*      boparams;
	switch (mo->type) {
		case map_object_prop:
			prparams = get_map_object_params(mo);
			prparams->sprite = sprite_game_over;
			prparams->z = 0;
			break;
		case map_object_killzone:
			kzparams = get_map_object_params(mo);
			kzparams->w = 16 << fbits;
			kzparams->h = 16 << fbits;
			kzparams->sprite = sprite_spike_a;
			break;
		case map_object_conttrig:
			ctparams = get_map_object_params(mo);
			ctparams->w = 16 << fbits;
			ctparams->h = 16 << fbits;
			strcpy(ctparams->level, "level1.lvl");
			break;
		case map_object_gamesaver:
			gsparams = get_map_object_params(mo);
			gsparams->w = 16 << fbits;
			gsparams->h = 16 << fbits;
			break;
		case map_object_pickup:
			piparams = get_map_object_params(mo);
			piparams->item = item_bomb;
			break;
		case map_object_cuttrig:
			cuparams = get_map_object_params(mo);
			cuparams->w = 16 << fbits;
			cuparams->h = 16 << fbits;
			strcpy(cuparams->cut, "cut_intro.cut");
			break;
		case map_object_boss:
			boparams = get_map_object_params(mo);
			boparams->type = boss_type_cia;
			break;
		default:
			break;
	}
}

void create_mo(World* world, Map_Object_Type type) {
	Map_Object* obj;
	char* buf;
	int size, i;
	size = get_map_object_size(type);
	obj = arena_alloc(&ed.mo_arena, size);
	buf = (char*)obj;
	for (i = 0; i < size; i++) {
		buf[i] = 0;
	}
	obj->type = type;
	init_new_map_object(obj);
	instance_map_object(world, obj);
}

int ed_load_level(World* world, const char* name) {
	Arena a;
	const cfg_Object* obj;
	char* src;
	FILE* file;
	int fsize;
	const int arena_size = level_loader_memory;
	init_arena(&a, galloc(arena_size), arena_size);
	file = fopen(name, "r");
	if (!file) return 0;
	fseek(file, 0, SEEK_END);
	fsize = ftell(file);
	rewind(file);
	src = arena_alloc(&a, fsize + 1);
	size = fread(src, 1, fsize, file);
	fclose(file);
	src[fsize] = 0;
	if (ed.mo_arena.buffer) {
		gfree(ed.mo_arena.buffer);
	}
	ed.mo_count = 0;
	init_arena(
		&ed.mo_arena,
		galloc(mo_arena_size),
		mo_arena_size
	);
	for (obj = cfg_parse(src, &a); obj; obj = obj->next) {
		if (!strcmp("level", obj->name)) {
			process_level_meta(world, obj);
		} else if (!strcmp("layer", obj->name)) {
			process_level_layer(world, obj);
		} else if (!strcmp("colliders", obj->name)) {
			process_level_colliders(world, obj);
		} else if (!strcmp("object", obj->name)) {
			process_level_object(world, obj);
		}
	}
	load_init(world, ed.lm.spx, ed.lm.spy);
	gfree(a.buffer);
	return 1;
}

int ed_save_cutscene(const char* name) {
	FILE* file;
	Cutscene* s;
	Cut_Cmd* cmd, * end;
	Cut_Cmd_Moveto* cmdmt;
	Cut_Cmd_Dialog* cmddi;
	const char* n;
	file = fopen(name, "w");
	if (!file) return 0;
	s = &ed.ct;
	fprintf(
		file,
		"[cutscene]\nendframe: %d\n",
		s->endframe
	);
	end = (Cut_Cmd*)(s->buf + s->buf_size);
	for (
		cmd = (Cut_Cmd*)s->buf;
		cmd != end;
		cmd = (Cut_Cmd*)((char*)cmd + cmd->size)
	) {
		switch (cmd->type) {
#define x(name_, com) \
			case cut_cmd_ ## name_: \
				n = #name_; \
				break;
			cut_cmd_type_xmacro()
#undef x
			default: break;
		}
		fprintf(
			file,
			"[%s]\nframe: %d\n",
			n,
			cmd->frame
		);
		switch (cmd->type) {
			case cut_cmd_moveto:
				cmdmt = (Cut_Cmd_Moveto*)cmd;
				fprintf(
					file,
					"handle: \"%s\n"
					"x: %d\ny: %d\n"
					"speed: %d\n",
					cmdmt->handle,
					cmdmt->x,
					cmdmt->y,
					cmdmt->speed
				);
				break;
			case cut_cmd_dialog:
				cmddi = (Cut_Cmd_Dialog*)cmd;
				fprintf(
					file,
					"face: \"%s\nsay: \"%s\n",
					get_sprite_name(cmddi->face),
					(char*)(cmddi + 1)
				);
				break;
			default: break;
		}
	}
	fclose(file);
	return 1;
}

int ed_load_cutscene(const char* name) {
	FILE* file;
	int fsize;
	char* src;
	const cfg_Object* obj;
	Arena a;
	const int arena_size = cutscene_loader_memory;
	init_arena(&a, galloc(arena_size), arena_size);
	if (ed.ct_arena.buffer) {
		gfree(ed.ct_arena.buffer);
	}
	init_arena(
		&ed.ct_arena,
		galloc(ct_arena_size),
		ct_arena_size
	);
	file = fopen(name, "r");
	if (!file) return 0;
	fseek(file, 0, SEEK_END);
	fsize = ftell(file);
	rewind(file);
	src = arena_alloc(&a, fsize + 1);
	size = fread(src, 1, fsize, file);
	fclose(file);
	src[fsize] = 0;
	init_cutscene(&ed.ct);
	obj = cfg_parse(src, &a);
	cutscene_load_cfg(&ed.ct, obj, &ed.ct_arena);
	gfree(a.buffer);
	return 1;
}

void real_edit(int* v, int step) {
	unsigned long p;
	mu_Real r, s;
	p = (unsigned long)v;
	r = (mu_Real)*v   / (mu_Real)(1 << fbits);
	s = (mu_Real)step / (mu_Real)(1 << fbits);
	mu_push_id(mu, &p, sizeof p);
	mu_number(mu, &r, s);
	mu_pop_id(mu);
	*v = (int)(r * (mu_Real)(1 << fbits));
}

void int_edit(int* v, int step) {
	unsigned long p;
	mu_Real r, s;
	p = (unsigned long)v;
	r = (mu_Real)*v;
	s = (mu_Real)step;
	mu_push_id(mu, &p, sizeof p);
	mu_number(mu, &r, s);
	mu_pop_id(mu);
	*v = (int)r;
}

void sprite_selector(int* s) {
	if (mu_button(mu, get_sprite_name(*s))) {
		mu_open_popup(mu, "select_prop_sprite");
	}
	if (mu_begin_popup(mu, "select_prop_sprite")) {
#define x(name) \
	if (mu_button(mu, #name) & MU_RES_SUBMIT) \
		*s = name;
		sprites_xmacro()
#undef x
		mu_end_popup(mu);
	}
}

void animation_selector(int* s) {
	if (mu_button(mu, get_animation_name(*s))) {
		mu_open_popup(mu, "select_prop_animation");
	}
	if (mu_begin_popup(mu, "select_prop_animation")) {
#define x(name, bitmap, frames, frame_count, slowness) \
	if (mu_button(mu, #name) & MU_RES_SUBMIT) \
		*s = name;
		animations_xmacro()
#undef x
		mu_end_popup(mu);
	}
}

void init_editor(void) {
	mu = stack_alloc(sizeof *mu);
	mu_init(mu);
	mu->style = &editor_style;
	mu->text_width  = text_width;
	mu->text_height = text_height;
	ed.tile = map_tile_air;
	ed.colmask = -1;
	ed.tool = editor_tool_paint;
	ed.show = 1;
	ed.showcols = 0;
	ed.showspawn = 1;
	ed.layer = map_layer_midground;
	ed.lm.spx = 0;
	ed.lm.spy = 0;
	ed.name[0] = 0;
	ed.lname[0] = 0;
	ed.cname[0] = 0;
	ed.ctcmd = 0;
	ed.ctcmd_frame = 0;
	ed.cname[0] = 0;
	ed.mo_arena.buffer = 0;
	ed.ct_arena.buffer = 0;
	ed.mo_count = 0;
	init_cutscene(&ed.ct);
	pause_gameplay = 1;
}

void deinit_editor(void) {

}

static void draw_axis(int x, int y) {
	Rectangle a, b;
	const Colour r = { 255, 64,  64,  200 };
	const Colour g = { 64,  255, 64,  200 };
	a.x = x;
	a.y = y;
	a.w = 16;
	a.h = 1;
	b.x = x;
	b.y = y;
	b.w = 1;
	b.h = 16;
	rc_add_cmd_rect(
		&a,
		r
	);
	rc_add_cmd_rect(
		&b,
		g
	);
}

static void draw_axisl(int x, int y, const char* label) {
	const Colour w = { 255, 255, 255, 200 };
	draw_axis(x, y);
	rc_add_cmd_rffont_text_col(
		get_smol_font(),
		x + 1,
		y + 1,
		label,
		w
	);
}

static void draw_spawn(void) {
	draw_axisl(
		ed.lm.spx >> fbits,
		ed.lm.spy >> fbits,
		"spawn"
	);
}


void draw_tile(void) {
	mu_Image img;
	const Sprite* s;
	const Bitmap* b;
	s = get_sprite(sprite_tile_box);
	b = get_bitmap(s->bitmap);
	img.w = b->w;
	img.h = b->h;
	img.pixels = b->pixels;
	mu_image_button_ex(mu, 0, &img, *(mu_Rect*)&s->rect,
	mu_color(255, 255, 255, 255));
}

void draw_tools(World* world) {
	mu_label(mu, "Tool:");
#define x(type, name) \
	if (mu_button(mu, name) & MU_RES_SUBMIT) { \
		ed.tool = type; \
	}
	editor_tool_xmacro
#undef x
	if (mu_button(mu, "Create Object") & MU_RES_SUBMIT) {
		mu_open_popup(mu, "create_object_popup");
	}
	mu_checkbox(mu, "Show Colliders", &ed.showcols);
	mu_checkbox(mu, "Show Spawn", &ed.showspawn);
	if (mu_begin_popup(mu, "create_object_popup")) {
#define x(type, factory, editor, spawner, params) \
	if (mu_button(mu, #type)) \
		create_mo(world, map_object_ ## type);
		map_objects_xmacro()
#undef x
		mu_end_popup(mu);
	}
}

void draw_tile_selector(void) {
	mu_Image img;
	const Bitmap* b;
	Rectangle rect;
#define width map_tile_size + 6
	int widths[] = { width, width, width };
#undef width
	mu_label(mu, "Tile:");
	mu_layout_begin_column(mu);
	mu_layout_row(mu, 3, widths, map_tile_size + 6);
	rect.w = rect.h = map_tile_size;
	b = get_bitmap(asset_id_map);
	img.w = b->w;
	img.h = b->h;
	img.pixels = b->pixels;
#define x(type, xo, yo) { \
	rect.x = xo; \
	rect.y = yo; \
	if (mu_image_button_ex( \
		mu, \
		0, \
		&img, \
		*(mu_Rect*)&rect, \
		mu_color(255, 255, 255, 255) \
	) & MU_RES_SUBMIT) { \
		ed.tile = type; \
	} \
}
	map_tiles_xmacro
#undef x
	mu_layout_end_column(mu);
}

void draw_collider_selector(void) {
	const Bitmap* b = get_bitmap(asset_id_colmask);
	mu_Image img;
	mu_Rect rect = {
		0,
		0,
		map_tile_size,
		map_tile_size
	};
	const int count = 1 << 4;
	int i;
#define width map_tile_size + 6
	int widths[] = { width, width, width };
#undef width
	img.w = b->w;
	img.h = b->h;
	img.pixels = b->pixels;
	mu_label(mu, "Collider Mask:");
	if (mu_button(mu, "None"))
		ed.colmask = -1;
	mu_layout_begin_column(mu);
	mu_layout_row(mu, 3, widths, map_tile_size + 6);
	for (i = 0; i < count; i++) {
		rect.x = i * map_tile_size;
		if (mu_image_button_ex(
			mu,
			0,
			&img,
			rect,
			mu_color(255, 255, 255, 255)
		) & MU_RES_SUBMIT)
			ed.colmask = i;
	}
	mu_layout_end_column(mu);
}

void draw_layer_selector(void) {
	int widths[] = { 32, -1 };
	mu_layout_begin_column(mu);
	mu_layout_row(mu, 2, widths, 0);
	mu_label(mu, "Layer:");
	mu_push_id(mu, "layer_select_btn", 16);
	if (mu_button(mu, layer_names[ed.layer])) {
		mu_pop_id(mu);
		mu_open_popup(mu, "layer_select_popup");
	} else {
		mu_pop_id(mu);
	}
	mu_layout_end_column(mu);
	if (mu_begin_popup(mu, "layer_select_popup")) {
#define x(type, name) \
	if (mu_button( \
		mu, \
		name \
	) & MU_RES_SUBMIT) { \
		ed.layer = type; \
	}
		map_layers_xmacro
#undef x
		mu_end_popup(mu);
	}
}

void update_paint_tool(World* world, int mx, int my) {
	int tx, ty;
	Rectangle r;
	Map_Layer* layer;
	if (mbtn_pressed(mbtn_left)) {
		tx = mx / map_tile_size;
		ty = my / map_tile_size;
		r.x = mx;
		r.y = my;
		r.w = map_tile_size;
		r.h = map_tile_size;
		layer = &world->map.layers[ed.layer];
		layer->tiles[tx + ty * world->map.w] = ed.tile;
		/* TODO: Only recache the tile that changed. */
		cache_map_layer(&world->map, layer);
		rc_invalidate(&r);
	}
}

void update_erase_tool(World* world, int mx, int my) {
	int tx, ty;
	Rectangle r;
	Map_Layer* layer;
	if (mbtn_pressed(mbtn_left)) {
		tx = mx / map_tile_size;
		ty = my / map_tile_size;
		r.x = mx;
		r.y = my;
		r.w = map_tile_size;
		r.h = map_tile_size;
		layer = &world->map.layers[ed.layer];
		layer->tiles[tx + ty * world->map.w] = map_tile_air;
		/* TODO: Only recache the tile that changed. */
		cache_map_layer(&world->map, layer);
		rc_invalidate(&r);
	}
}

void draw_collider(unsigned char c, int x, int y) {
	Rectangle rect;
	Colour col = { 255, 255, 255, 100 };
	const Bitmap* mb;
	x *= map_tile_size;
	y *= map_tile_size;
	rect.w = map_tile_size;
	rect.h = map_tile_size;
	switch (get_collider_type(c)) {
		case map_collider_block:
			rect.x = x;
			rect.y = y;
			rc_add_cmd_rect(&rect, col);
			break;
		case map_collider_mask:
			mb = get_bitmap(asset_id_colmask);
			rect.x = (c & 0xf) * map_tile_size;
			rect.y = 0;
			rc_add_cmd_bmp_col(mb, x, y, &rect, col);
			break;
		default:
			break;
	}
}

void draw_colliders(World* world) {
	int x, y;
	Map* map = &world->map;
	unsigned char* col;
	col = map->colliders;
	for (y = 0; y < map->h; y++) {
		for (x = 0; x < map->w; x++) {
			if (*col)
				draw_collider(*col, x, y);
			col++;
		}
	}
}

void update_collider_tool(World* world, int mx, int my) {
	int x, y;
	Map* map = &world->map;
	unsigned char* col;
	ed.showcols = 1;
	x = mx / map_tile_size;
	y = my / map_tile_size;
	col = &map->colliders[x + y * map->w];
	if (mbtn_pressed(mbtn_left)) {
		if (ed.colmask == -1)
			*col = map_collider_block << 4;
		else
			*col = (map_collider_mask << 4) | ed.colmask;
	}
	if (mbtn_pressed(mbtn_right)) {
		*col = 0;
	}
}

void default_object_editor(
	World* world,
	Map_Object* obj,
	Entity e
) {
	int widths[] = { 25, -1 };
	(void)e;
	mu_layout_begin_column(mu);
	mu_layout_row(mu, 2, widths, 0);
	mu_label(mu, "Type:");
	mu_label(mu, get_map_object_name(obj->type));
	mu_layout_end_column(mu);
	if (mu_button(mu, "Delete") & MU_RES_SUBMIT) {
		destroy_mo(world, ed.selected_mo);
		ed.dragging_mo = 0;
		ed.selected_mo = 0;
	}
	if (mu_button(mu, "Duplicate") & MU_RES_SUBMIT) {
		dup_mo(world, ed.selected_mo);
	}
}

void draw_object_editor(World* world) {
	Map_Object_Editor f;
	Map_Object_Instance* i;
	int j;
	if (mu_begin_window_ex(
		mu,
		"Object",
		mu_rect(119, 174, 98, 64),
		MU_OPT_NOCLOSE
	)) {
		for (j = 0; j < ed.mo_count; j++) {
			i = &ed.mos[j];
			f = mo_editors[i->obj->type];
			f(world, i->obj, i->e);
		}
		mu_end_window(mu);
	}
}

void update_picker_tool(World* world, int mx, int my) {
	int i, eid, h = 0;
	Com_Bitset mask;
	const Com_Bitset pickmask = ctype_position;
	const Colour rcol = { 255, 0, 0, 128 };
	const Colour ccol = { 255, 0, 255, 100 };
	Entity e;
	Rectangle rect;
	Map_Object_Instance* mo;
	CSprite* spr;
	CPosition* pos;
	CCollider* col;
	for (i = 0; i < ed.mo_count; i++) {
		mo = &ed.mos[i];
		e = mo->e;
		eid = entity_index(e);
		mask = world->bitset[eid] & pickmask;
		if (mask == pickmask) {
			mask = world->bitset[eid];
			pos = &world->positions[eid];
			if (mask & ctype_sprite) {
				spr = &world->sprites[eid];
				rect.x = pos->x >> fbits;
				rect.y = pos->y >> fbits;
				rect.w = spr->rect.w;
				rect.h = spr->rect.h;
			} else if (mask & ctype_collider) {
				col = &world->colliders[eid];
				rect.x = (col->x + pos->x) >> fbits;
				rect.y = (col->y + pos->y) >> fbits;
				rect.w = col->w >> fbits;
				rect.h = col->h >> fbits;
			} else continue;
			if (point_rect_overlap(&rect, mx, my)) {
				h = 1;
				if (mbtn_just_pressed(mbtn_left))
					ed.selected_mo = mo;
			}
			if (ed.selected_mo == mo) {
				rc_add_cmd_rect(&rect, rcol);
				draw_axis(rect.x, rect.y);
				if (point_rect_overlap(&rect, mx, my)) {
					if (mbtn_just_pressed(mbtn_left)) {
						ed.dragging_mo = mo;
						ed.doffx = rect.x - mx;
						ed.doffy = rect.y - my;
					}
				}
			}
			if (mask & ctype_collider) {
				col = &world->colliders[eid];
				rect.x = (col->x + pos->x) >> fbits;
				rect.y = (col->y + pos->y) >> fbits;
				rect.w = col->w >> fbits;
				rect.h = col->h >> fbits;
				rc_add_cmd_rect(&rect, ccol);
			}
		}
	}
	if (mbtn_just_released(mbtn_left)) {
		ed.dragging_mo = 0;
		if (!h)
			ed.selected_mo = 0;
	}
	if (ed.dragging_mo) {
		mo = ed.dragging_mo;
		eid = entity_index(mo->e);
		pos = &world->positions[eid];
		pos->x = (mx + ed.doffx) << fbits;
		pos->y = (my + ed.doffy) << fbits;
	}
}

void update_editor_tool(World* world) {
	int mx, my;
	mx = get_mouse_x();
	my = get_mouse_y();
	switch (ed.tool) {
		case editor_tool_paint:
			update_paint_tool(world, mx, my);
			break;
		case editor_tool_erase:
			update_erase_tool(world, mx, my);
			break;
		case editor_tool_collider:
			update_collider_tool(world, mx, my);
			break;
		case editor_tool_picker:
			update_picker_tool(world, mx, my);
			break;
		default:
			break;
	}
}

void draw_file(World* world) {
	char name[64];
	if (mu_button(mu, "Level file"))
		mu_open_popup(mu, "level_file_popup");
	if (mu_button(mu, "Cutscene file"))
		mu_open_popup(mu, "cutscene_file_popup");
	if (mu_begin_popup(mu, "level_file_popup")) {
		mu_textbox(mu, ed.name, sizeof ed.name);
		strcpy(name, "intermediate/");
		strcat(name, ed.name);
		strcat(name, ".txt");
		if (mu_button(mu, "Load") & MU_RES_SUBMIT)
			ed_load_level(world, name);
		if (mu_button(mu, "Save") & MU_RES_SUBMIT)
			ed_save_level(world, name);
		mu_end_popup(mu);
	}
	if (mu_begin_popup(mu, "cutscene_file_popup")) {
		mu_textbox(mu, ed.cname, sizeof ed.cname);
		strcpy(name, "intermediate/");
		strcat(name, ed.cname);
		strcat(name, ".txt");
		if (mu_button(mu, "Load") & MU_RES_SUBMIT)
			ed_load_cutscene(name);
		if (mu_button(mu, "Save") & MU_RES_SUBMIT)
			ed_save_cutscene(name);
		mu_end_popup(mu);
	}
}

void draw_metadata(World* world) {
	int widths[] = { 45, -1 };
	(void)world;
	mu_label(mu, "Metadata:");
	mu_layout_begin_column(mu);
	mu_layout_row(mu, 2, widths, 0);
	mu_label(mu, "Name: "); mu_textbox(mu, ed.lname, sizeof ed.lname);
	mu_label(mu, "Spawn X:"); real_edit(&ed.lm.spx, f1);
	mu_label(mu, "Spawn Y:"); real_edit(&ed.lm.spy, f1);
	mu_layout_end_column(mu);
}

void draw_timeline(void) {
	int i, ch;
	int x, y, mos;
	Cut_Cmd* cmd, * end;
	Cutscene* s;
	mu_Rect r, h, mr, str;
	mu_Color
		a  = { 0,   0,   100, 255 },
		b  = { 25,  25,  25,  255 },
		o  = { 0,   0,   255, 255 },
		o2 = { 0,   255, 0,   255 },
		m  = { 255, 0,   0,   255 };
	mr.x = 0;
	mr.y = 0;
	mr.w = cutscene_editor_timeline_el_w * ed.ct.endframe;
	mr.h = cutscene_editor_timeline_el_h;
	mu_layout_set_next(mu, mr, 1);
	r = mu_layout_next(mu);
	str = r;
	str.h *= cutscene_editor_channels;
	y = r.y;
	s = &ed.ct;
	for (i = 0; i < cutscene_editor_channels; i++) {
		mu_draw_rect(mu, r, i % 2? a: b);
		r.y += r.h;
	}
	end = (Cut_Cmd*)(s->buf + s->buf_size);
	mos = mu_mouse_over(mu, str);
	if (
		ed.ctcmd &&
		mbtn_just_pressed(mbtn_left) &&
		mos
	) ed.ctcmd = 0;
	/* todo figure out something faster here */
	for (i = 0; i < ed.ct.endframe; i++) {
		ch = 0;
		for (
			cmd = (Cut_Cmd*)s->buf;
			cmd != end;
			cmd = (Cut_Cmd*)((char*)cmd + cmd->size)
		) {
			if (cmd->frame != i) continue;
			h.x =
				str.x +
				cmd->frame *
				cutscene_editor_timeline_el_w;
			h.y =
				str.y +
				ch *
				cutscene_editor_timeline_el_h;
			h.w = cutscene_editor_timeline_el_w;
			h.h = cutscene_editor_timeline_el_h;
			if (
				mu_mouse_over(mu, h) &&
				mbtn_just_pressed(mbtn_left)
			) {
				ed.ctcmd = cmd;
			}
			mu_draw_rect(
				mu,
				h,
				cmd == ed.ctcmd? o2: o
			);
			h.x++;
			h.y++;
			h.w -= 2;
			h.h -= 2;
			mu_draw_rect(
				mu,
				h,
				m
			);
			ch++;
		}
	}
	if (
		ed.ctcmd &&
		mbtn_pressed(mbtn_left) &&
		mos
	) {
		x = clamp(get_mouse_x() - r.x, 0, r.w - 1);
		y = clamp(get_mouse_y() - r.y, 0, r.h - 1);
		x /= cutscene_editor_timeline_el_w;
		y /= cutscene_editor_timeline_el_h;
		ed.ctcmd->frame = x;
	}
}

void resize_cmd(Cut_Cmd* cmd, int ns) {
	int i, o, d, e;
	Cutscene* s;
	ns = (ns + 7) & -8;
	s = &ed.ct;
	o = ((char*)cmd - (char*)s->buf);
	d = ns - cmd->size;
	if (d) {
		memmove(
			(char*)cmd + ns,
			(char*)cmd + cmd->size,
			s->buf_size - o - cmd->size
		);
		e = o + ns;
		for (i = o + cmd->size; i < e; i++)
			s->buf[i] = 0;
		cmd->size = ns;
		s->buf_size += d;
	}
}

void delete_cmd(Cut_Cmd* cmd) {
	Cutscene* s = &ed.ct;
	int size = cmd->size;
	int o = ((char*)cmd - (char*)s->buf);
	memmove(
		(char*)cmd,
		(char*)cmd + cmd->size,
		s->buf_size - o - size
	);
	s->buf_size -= size;
	if (cmd == ed.ctcmd) ed.ctcmd = 0;
}

void draw_ctcmd_moveto(Cut_Cmd_Moveto* cmd) {
	int widths[] = { 32, -1 };
	mu_layout_begin_column(mu);
	mu_layout_row(mu, 2, widths, 0);
	mu_label(mu, "X:"); real_edit(&cmd->x, f1);
	mu_label(mu, "Y:"); real_edit(&cmd->y, f1);
	mu_label(mu, "Speed:"); real_edit(&cmd->speed, 1);
	mu_label(mu, "Handle:");
	mu_textbox(mu, cmd->handle, sizeof cmd->handle);
	mu_layout_end_column(mu);
	draw_axisl(cmd->x >> fbits, cmd->y >> fbits, cmd->handle);
}

void draw_ctcmd_dialog(Cut_Cmd_Dialog* cmd) {
	int widths[] = { 32, -1 };
	char buf[256];
	strcpy(buf, (char*)(cmd + 1));
	mu_layout_begin_column(mu);
	mu_layout_row(mu, 2, widths, 0);
	mu_label(mu, "Face:"); 
	sprite_selector(&cmd->face);
	mu_label(mu, "Say:"); 
	mu_textbox(mu, buf, sizeof buf);
	resize_cmd(
		(Cut_Cmd*)cmd,
		strlen(buf) + 1 + sizeof *cmd
	);
	strcpy((char*)(cmd + 1), buf);
	mu_layout_end_column(mu);
}

int ctcmd_default_size(Cut_Cmd_Type t) {
	switch (t) {
		case cut_cmd_moveto:
			return sizeof(Cut_Cmd_Moveto);
		case cut_cmd_dialog:
			return sizeof(Cut_Cmd_Dialog) + 1;
		default:
			print_err("Invalid cutscene command type.");
			return 0;
	}
}

void draw_ctcmd(Cut_Cmd* cmd) {
	int widths[] = { 32, -1 };
	int s;
	if (mu_button(mu, "Delete") & MU_RES_SUBMIT)
		delete_cmd(cmd);
	mu_layout_begin_column(mu);
	mu_layout_row(mu, 2, widths, 0);
	mu_label(mu, "Type:");
	if (mu_button(mu,
		get_cut_cmd_type_name(cmd->type)
	)) {
		mu_open_popup(mu, "select_cmd_type");
	}
	if (mu_begin_popup(mu, "select_cmd_type")) {
#define x(name, com) \
	if ( \
		mu_button(mu, #name) & MU_RES_SUBMIT && \
		cmd->type != cut_cmd_ ## name \
	) { \
		s = ctcmd_default_size(cut_cmd_ ## name); \
		resize_cmd(cmd, s); \
		memset(cmd + 1, 0, s - sizeof *cmd); \
		cmd->type = cut_cmd_ ## name; \
	}
		cut_cmd_type_xmacro()
#undef x
		mu_end_popup(mu);
	}
	mu_label(mu, "Frame:"); int_edit(&cmd->frame, 1);
	mu_layout_end_column(mu);
	switch (cmd->type) {
		case cut_cmd_moveto:
			draw_ctcmd_moveto(
				(Cut_Cmd_Moveto*)cmd
			);
			break;
		case cut_cmd_dialog:
			draw_ctcmd_dialog(
				(Cut_Cmd_Dialog*)cmd
			);
			break;
		default: break;
	}
}

void create_cmd(Cut_Cmd_Type t, int frame) {
	int s;
	Cut_Cmd* c;
	s = ctcmd_default_size(t);
	c = alloc_cmd(&ed.ct, s);
	if (!c) return;
	c->type = t;
	c->frame = frame;
}

void draw_cutscene_editor(void) {
	int widths[] = { 45, -1 };
	if (mu_begin_window_ex(
		mu,
		"Cutscene",
		mu_rect(renderer_w - 98 - 2, 141, 98, 97),
		MU_OPT_CLOSED
	)) {
		mu_layout_begin_column(mu);
		mu_layout_row(mu, 2, widths, 0);
		mu_label(mu, "Frames:"); int_edit(&ed.ct.endframe, 1);
		mu_layout_end_column(mu);
		ed.ct.endframe = maxi(0, ed.ct.endframe);
		if (mu_button(mu, "Create Command"))
			mu_open_popup(mu, "create_cmd_popup");
		if (mu_begin_popup(mu, "create_cmd_popup")) {
			mu_layout_begin_column(mu);
			mu_layout_row(mu, 2, widths, 0);
			mu_label(mu, "Frame:");
			int_edit(&ed.ctcmd_frame, 1);
			mu_layout_end_column(mu);
	#define x(name, com) \
		if (mu_button(mu, #name) & MU_RES_SUBMIT) \
			create_cmd(cut_cmd_ ## name, ed.ctcmd_frame);
			cut_cmd_type_xmacro()
	#undef x
			mu_end_popup(mu);
		}
		mu_end_window(mu);
	}
	if (mu_begin_window_ex(
		mu,
		"Timeline",
		mu_rect(2, 174, 215, 64),
		MU_OPT_CLOSED
	)) {
		draw_timeline();
		mu_end_window(mu);
	}
	if (ed.ctcmd) {
		if (mu_begin_window(
			mu,
			"Command",
			mu_rect(renderer_w - 98 - 2, 2, 98, 136)
		)) {
			draw_ctcmd(ed.ctcmd);
			mu_end_window(mu);
		}
	}
}

void update_editor(World* world) {
	mu_begin(mu);
	if (btn_just_pressed(btn_pc_f12)) {
		ed.show = !ed.show;
		if (ed.show) {
			pause_gameplay = ed.show;
		}
	}
	if (ed.show) {
		if (mu_begin_window_ex(
			mu,
			"Toolbox",
			mu_rect(2, 2, 0, 0),
			MU_OPT_NOCLOSE | MU_OPT_AUTOSIZE | MU_OPT_NORESIZE
		)) {
			if (mu_button(mu, "Edit level"))
				mu_open_window(mu, "Level");
			if (mu_button(mu, "Edit cutscene")) {
				mu_open_window(mu, "Cutscene");
				mu_open_window(mu, "Timeline");
			}
			mu_checkbox(mu, "Pause Game", &pause_gameplay);
			draw_file(world);

			if (mu_begin_window_ex(
				mu,
				"Level",
				mu_rect(renderer_w - 98 - 2, 2, 98, 236),
				MU_OPT_CLOSED
			)) {
				draw_metadata(world);
				draw_tools(world);
				draw_layer_selector();
				draw_tile_selector();
				draw_collider_selector();
				mu_end_window(mu);

				if (!mu->hover_root && !mu->focus && !mu->hover) {
					update_editor_tool(world);
				}
			}
			draw_cutscene_editor();
			mu_end_window(mu);
		}
		if (ed.selected_mo)
			draw_object_editor(world);
	}
	if (ed.showcols) {
		draw_colliders(world);
	}
	if (ed.showspawn) {
		draw_spawn();
	}
	mu_end(mu);
}

void render_editor() {
	mu_Command* cmd = 0;
	const Sprite* sprite;
	int i, tl;
	const char* tin;
	Bitmap bmp;
	mu_input_mousemove(mu, get_mouse_x(), get_mouse_y());
	mu_input_scroll(
		mu,
		get_mscroll_x() * 4,
		get_mscroll_y() * 4
	);
	for (i = 0; i < mbtn_count; i++) {
		if (mbtn_just_pressed(i)) {
			mu_input_mousedown(
				mu,
				get_mouse_x(),
				get_mouse_y(),
				1 << i
			);
		}
		if (mbtn_just_released(i)) {
			mu_input_mouseup(
				mu,
				get_mouse_x(),
				get_mouse_y(),
				1 << i
			);
		}
	}
	for (i = btn_pc_shift; i <= btn_pc_return; i++) {
		if (btn_just_pressed(i)) {
			mu_input_keydown(mu, 1 << (i - btn_pc_shift));
		}
		if (btn_just_released(i)) {
			mu_input_keyup(mu, 1 << (i - btn_pc_shift));
		}
	}
	tin = get_text_input(&tl);
	if (tin) {
		mu_input_text(mu, tin);
	}
	while (mu_next_command(mu, &cmd)) {
		switch (cmd->type) {
			case MU_COMMAND_RECT:
				rc_add_cmd_rect(
					(Rectangle*)&cmd->rect.rect,
					*(Colour*)&cmd->rect.color
				);
				break;
			case MU_COMMAND_TEXT:
				rc_add_cmd_rffont_text_col(
					get_smol_font(),
					cmd->text.pos.x,
					cmd->text.pos.y,
					cmd->text.str,
					*(Colour*)&cmd->text.color
				);
				break;
			case MU_COMMAND_CLIP:
				cmd->clip.rect.w += cmd->clip.rect.x;
				cmd->clip.rect.h += cmd->clip.rect.y;
				rc_add_cmd_clip((Rectangle*)&cmd->clip.rect);
				break;
			case MU_COMMAND_ICON:
				sprite = get_sprite(
					cmd->icon.id +
					sprite_icon_close -
					1
				);
				rc_add_cmd_bmp_cp_col(
					get_bitmap(sprite->bitmap),
					cmd->icon.rect.x + 1,
					cmd->icon.rect.y + 1,
					&sprite->rect,
					*(Colour*)&cmd->icon.color
				);
				break;
			case MU_COMMAND_IMAGE:
				bmp.pixels = (Colour*)cmd->image.img.pixels;
				bmp.w = cmd->image.img.w;
				bmp.h = cmd->image.img.h;
				rc_add_cmd_bmp_cp_col(
					&bmp,
					cmd->image.x,
					cmd->image.y,
					(Rectangle*)&cmd->image.rect,
					*(Colour*)&cmd->image.color
				);
				break;
			default: break;
		}
	}
}

/* Map object editors */
void edit_prop(
	World* world,
	Map_Object* obj,
	Entity e
) {
	int eid;
	int widths[] = { 45, -1 };
	prop_paramstruct* params;
	CPosition* pos;
	CSprite* spr;
	CAnimated* anim;
	eid = entity_index(e);
	params = get_map_object_params(obj);
	pos = &world->positions[eid];
	spr = &world->sprites[eid];
	anim = &world->animated[eid];
	params->x = pos->x;
	params->y = pos->y;
	params->z = spr->z;

	if (ed.selected_mo && ed.selected_mo->e == e) {
		default_object_editor(world, obj, e);
		if (mu_checkbox(
				mu,
				"Animated:",
				&params->anim
		) & MU_RES_CHANGE) {
			params->sprite = 0;
			if (params->anim) {
				add_components(
					world,
					e,
					ctype_animated
				);
			} else {
				remove_components(
					world,
					e,
					ctype_animated
				);
			}
		}
		mu_checkbox(mu, "Destructable:", &params->destructable);
		mu_checkbox(mu, "Collidable:", &params->collidable);
		if (params->anim) {
			mu_checkbox(mu, "Random start:", &params->random);
			mu_layout_begin_column(mu);
			mu_layout_row(mu, 2, widths, 0);
			mu_label(mu, "Animation:");
			animation_selector(&params->sprite);
		} else {
			mu_layout_begin_column(mu);
			mu_layout_row(mu, 2, widths, 0);
			mu_label(mu, "Sprite:");
			sprite_selector(&params->sprite);
		}
		mu_label(mu, "X:"); real_edit(&params->x, f1);
		mu_label(mu, "Y:"); real_edit(&params->y, f1);
		mu_label(mu, "Z:"); int_edit(&params->z, 1);
		mu_layout_end_column(mu);
		pos->x = params->x;
		pos->y = params->y;
		if (params->anim) {
			anim->id = params->sprite;
			anim->frame = 0;
			anim->timer = 0;
		} else
			init_csprite(spr, params->sprite);
		spr->z = params->z;
	}
}

void prop_saver(const Map_Object* obj, FILE* file) {
	const prop_paramstruct* params;
	params = get_map_object_params(obj);
	if (params->anim) {
		fprintf(
			file,
			"animation: \"%s\n",
			get_animation_name(params->sprite)
		);
		fprintf(
			file,
			"randomise_start: %d\n",
			params->random
		);
	} else
		fprintf(
			file,
			"sprite: \"%s\n",
			get_sprite_name(params->sprite)
		);
	fprintf(file, "x: %d\n", params->x);
	fprintf(file, "y: %d\n", params->y);
	fprintf(file, "z: %d\n", params->z);
	if (params->collidable)
		fprintf(file, "collidable: 1\n");
	if (params->destructable)
		fprintf(file, "destructable: 1\n");
}

void edit_cop(
	World* world,
	Map_Object* obj,
	Entity e
) {
	int eid;
	int widths[] = { 45, -1 };
	cop_paramstruct* params;
	CPosition* pos;
	eid = entity_index(e);
	params = get_map_object_params(obj);
	pos = &world->positions[eid];
	params->x = pos->x;
	params->y = pos->y;

	if (ed.selected_mo && ed.selected_mo->e == e) {
		default_object_editor(world, obj, e);
		mu_layout_begin_column(mu);
		mu_layout_row(mu, 2, widths, 0);
		mu_label(mu, "X:"); real_edit(&params->x, f1);
		mu_label(mu, "Y:"); real_edit(&params->y, f1);
		mu_layout_end_column(mu);
		pos->x = params->x;
		pos->y = params->y;
	}
}

void cop_saver(const Map_Object* obj, FILE* file) {
	const cop_paramstruct* params;
	params = get_map_object_params(obj);
	fprintf(file, "x: %d\n", params->x);
	fprintf(file, "y: %d\n", params->y);
}

void edit_killzone(
	World* world,
	Map_Object* obj,
	Entity e
) {
	int eid;
	int widths[] = { 45, -1 };
	killzone_paramstruct* params;
	CPosition* pos;
	CCollider* col;
	CSprite* spr;
	eid = entity_index(e);
	params = get_map_object_params(obj);
	pos = &world->positions[eid];
	col = &world->colliders[eid];
	spr = &world->sprites[eid];
	params->x  = pos->x;
	params->y  = pos->y;
	params->ox = col->x;
	params->oy = col->y;
	params->w  = col->w;
	params->h  = col->h;
	if (ed.selected_mo && ed.selected_mo->e == e) {
		default_object_editor(world, obj, e);
		if (mu_button(mu, get_sprite_name(params->sprite))) {
			mu_open_popup(mu, "select_killzone_sprite");
		}
		if (mu_begin_popup(mu, "select_killzone_sprite")) {
#define x(name) \
	if (mu_button(mu, #name) & MU_RES_SUBMIT) \
		params->sprite = name;
			sprites_xmacro()
#undef x
			mu_end_popup(mu);
		}
		mu_layout_begin_column(mu);
		mu_layout_row(mu, 2, widths, 0);
		mu_label(mu, "X:");  real_edit(&params->x,  f1);
		mu_label(mu, "Y:");  real_edit(&params->y,  f1);
		mu_label(mu, "OX:"); real_edit(&params->ox, f1);
		mu_label(mu, "OY:"); real_edit(&params->oy, f1);
		mu_label(mu, "W:");  real_edit(&params->w,  f1);
		mu_label(mu, "H:");  real_edit(&params->h,  f1);
		mu_layout_end_column(mu);
		pos->x = params->x;
		pos->y = params->y;
		col->x = params->ox;
		col->y = params->oy;
		col->w = params->w;
		col->h = params->h;
		init_csprite(spr, params->sprite);
	}
}

void killzone_saver(const Map_Object* obj, FILE* file) {
	const killzone_paramstruct* params;
	params = get_map_object_params(obj);
	fprintf(
		file,
		"sprite: \"%s\n",
		get_sprite_name(params->sprite)
	);
	fprintf(file, "x: %d\n", params->x);
	fprintf(file, "y: %d\n", params->y);
	fprintf(file, "ox: %d\n", params->ox);
	fprintf(file, "oy: %d\n", params->oy);
	fprintf(file, "w: %d\n", params->w);
	fprintf(file, "h: %d\n", params->h);
}

void edit_conttrig(
	World* world,
	Map_Object* obj,
	Entity e
) {
	int eid;
	int widths[] = { 45, -1 };
	conttrig_paramstruct* params;
	CPosition* pos;
	CCollider* col;
	CConttrig* conttrig;
	eid = entity_index(e);
	params = get_map_object_params(obj);
	pos = &world->positions[eid];
	col = &world->colliders[eid];
	conttrig = &world->conttrigs[eid];
	params->x = pos->x;
	params->y = pos->y;
	params->w = col->w;
	params->h = col->h;
	strcpy(params->level, conttrig->level);
	if (ed.selected_mo && ed.selected_mo->e == e) {
		default_object_editor(world, obj, e);
		mu_layout_begin_column(mu);
		mu_layout_row(mu, 2, widths, 0);
		mu_label(mu, "X:"); real_edit(&params->x, f1);
		mu_label(mu, "Y:"); real_edit(&params->y, f1);
		mu_label(mu, "W:"); real_edit(&params->w, f1);
		mu_label(mu, "H:"); real_edit(&params->h, f1);
		mu_label(mu, "Level:"); mu_textbox(
			mu,
			params->level,
			sizeof params->level
		);
		mu_layout_end_column(mu);
		pos->x = params->x;
		pos->y = params->y;
		col->w = params->w;
		col->h = params->h;
		strcpy(conttrig->level, params->level);
	}
}

void conttrig_saver(const Map_Object* obj, FILE* file) {
	const conttrig_paramstruct* params;
	params = get_map_object_params(obj);
	fprintf(file, "x: %d\n", params->x);
	fprintf(file, "y: %d\n", params->y);
	fprintf(file, "w: %d\n", params->w);
	fprintf(file, "h: %d\n", params->h);
	fprintf(file, "level: \"%s\n", params->level);
}

void edit_gamesaver(
	World* world,
	Map_Object* obj,
	Entity e
) {
	int eid;
	int widths[] = { 45, -1 };
	gamesaver_paramstruct* params;
	CPosition* pos;
	eid = entity_index(e);
	params = get_map_object_params(obj);
	pos = &world->positions[eid];
	params->x = pos->x;
	params->y = pos->y;

	if (ed.selected_mo && ed.selected_mo->e == e) {
		default_object_editor(world, obj, e);
		mu_layout_begin_column(mu);
		mu_layout_row(mu, 2, widths, 0);
		mu_label(mu, "X:"); real_edit(&params->x, f1);
		mu_label(mu, "Y:"); real_edit(&params->y, f1);
		mu_label(mu, "W:"); real_edit(&params->w, f1);
		mu_label(mu, "H:"); real_edit(&params->h, f1);
		mu_layout_end_column(mu);
		pos->x = params->x;
		pos->y = params->y;
	}
}

void gamesaver_saver(
	const Map_Object* obj,
	FILE* file
) {
	const gamesaver_paramstruct* params;
	params = get_map_object_params(obj);
	fprintf(file, "x: %d\n", params->x);
	fprintf(file, "y: %d\n", params->y);
	fprintf(file, "w: %d\n", params->w);
	fprintf(file, "h: %d\n", params->h);
}

void edit_pickup(
	World* world,
	Map_Object* obj,
	Entity e
) {
	int eid;
	int widths[] = { 45, -1 };
	pickup_paramstruct* params;
	CPosition* pos;
	eid = entity_index(e);
	params = get_map_object_params(obj);
	pos = &world->positions[eid];
	params->x = pos->x;
	params->y = pos->y;

	if (ed.selected_mo && ed.selected_mo->e == e) {
		default_object_editor(world, obj, e);
		mu_layout_begin_column(mu);
		mu_layout_row(mu, 2, widths, 0);
		mu_label(mu, "X:"); real_edit(&params->x, f1);
		mu_label(mu, "Y:"); real_edit(&params->y, f1);
		mu_layout_end_column(mu);
		pos->x = params->x;
		pos->y = params->y;
	}
}

void pickup_saver(
	const Map_Object* obj,
	FILE* file
) {
	const pickup_paramstruct* params;
	const char* name;
	params = get_map_object_params(obj);
	name = get_item_name(params->item);
	fprintf(file, "x: %d\n", params->x);
	fprintf(file, "y: %d\n", params->y);
	fprintf(file, "item: %s\n", name);
}

void edit_cuttrig(
	World* world,
	Map_Object* obj,
	Entity e
) {
	int eid;
	int widths[] = { 45, -1 };
	cuttrig_paramstruct* params;
	CPosition* pos;
	CCollider* col;
	CCuttrig* cuttrig;
	eid = entity_index(e);
	params = get_map_object_params(obj);
	pos = &world->positions[eid];
	col = &world->colliders[eid];
	cuttrig = &world->cuttrigs[eid];
	params->x = pos->x;
	params->y = pos->y;
	params->w = col->w;
	params->h = col->h;
	strcpy(params->cut, cuttrig->cut);
	if (ed.selected_mo && ed.selected_mo->e == e) {
		default_object_editor(world, obj, e);
		mu_layout_begin_column(mu);
		mu_layout_row(mu, 2, widths, 0);
		mu_label(mu, "X:"); real_edit(&params->x, f1);
		mu_label(mu, "Y:"); real_edit(&params->y, f1);
		mu_label(mu, "W:"); real_edit(&params->w, f1);
		mu_label(mu, "H:"); real_edit(&params->h, f1);
		mu_label(mu, "Scene:"); mu_textbox(
			mu,
			params->cut,
			sizeof params->cut
		);
		mu_layout_end_column(mu);
		pos->x = params->x;
		pos->y = params->y;
		col->w = params->w;
		col->h = params->h;
		strcpy(cuttrig->cut, params->cut);
	}
}

void cuttrig_saver(const Map_Object* obj, FILE* file) {
	const cuttrig_paramstruct* params;
	params = get_map_object_params(obj);
	fprintf(file, "x: %d\n", params->x);
	fprintf(file, "y: %d\n", params->y);
	fprintf(file, "w: %d\n", params->w);
	fprintf(file, "h: %d\n", params->h);
	fprintf(file, "scene: \"%s\n", params->cut);
}

void edit_boss(
	World* world,
	Map_Object* obj,
	Entity e
) {
	int eid;
	int widths[] = { 45, -1 };
	boss_paramstruct* params;
	CPosition* pos;
	eid = entity_index(e);
	params = get_map_object_params(obj);
	pos = &world->positions[eid];
	params->x = pos->x;
	params->y = pos->y;

	if (ed.selected_mo && ed.selected_mo->e == e) {
		default_object_editor(world, obj, e);
		mu_layout_begin_column(mu);
		mu_layout_row(mu, 2, widths, 0);
		mu_label(mu, "X:"); real_edit(&params->x, f1);
		mu_label(mu, "Y:"); real_edit(&params->y, f1);
		mu_layout_end_column(mu);
		pos->x = params->x;
		pos->y = params->y;
	}
}

void boss_saver(const Map_Object* obj, FILE* file) {
	const boss_paramstruct* params;
	params = get_map_object_params(obj);
	fprintf(file, "x: %d\n", params->x);
	fprintf(file, "y: %d\n", params->y);
	fprintf(
		file,
		"boss_type: \"%s\n",
		get_boss_type_name(params->type)
	);
}

#endif
