#include <string.h>
#include "asset.h"
#include "map.h"
#include "profile.h"
#include "rcache.h"
#include "rect.h"
#ifdef DEBUG
#include "plat.h"
#endif

typedef struct {
	int x, y;
}  Tile_Offset;

static const Tile_Offset tile_offsets[] = {
#define x(n, ox, oy) { ox, oy },
	map_tiles_xmacro
#undef x
};

void init_map(Map* map, int w, int h) {
	int e, i;
	Colour* c;
	map->w = w;
	map->h = h;
	map->corpse_count = 0;
	e = map->w * map->h;
	for (i = 0; i < e; i++)
		map->colliders[i] = map_collider_null;
	e *= map_tile_size * map_tile_size;
	for (i = 0; i < e; i++) {
		c = &map->blood[i];
		c->r = 0;
		c->g = 0;
		c->b = 0;
		c->a = 0;
	}
	for (i = 0; i < map_layer_count; i++) {
		init_map_layer(map, &map->layers[i]);
	}
}

static void gen_background(Map* map, Bitmap* dst) {
	int x, y, w, h, c;
	unsigned seed;
	const Colour sky = { 20, 20, 255, 255 };
	Sprite_ID sid;
	const Sprite* s;
	w = map->w * map_tile_size;
	h = map->h * map_tile_size;
	for (y = 0; y < h; y++)
		for (x = 0; x < w; x++)
			dst->pixels[x + y * w] = sky;
	seed = get_seed();
	seed_rng(548944);
	c = rand_range(100, 500);
	for (x = 0; x < c; x++) {
		sid = (Sprite_ID)rand_range(
			sprite_star_blue_a,
			sprite_star_yellow_c
		);
		s = get_sprite(sid);
		rcopy(
			dst,
			get_bitmap(s->bitmap),
			rand_range(0, w),
			rand_range(0, h),
			&s->rect
		);
	}
	seed_rng(seed);
}

void cache_map(Map* map) {
	int i;
	const Corpse* corpse;
	const Sprite* sprite;
	const Bitmap* src;
	Bitmap dst, bak;
	dst.w = map->w * map_tile_size;
	dst.h = map->h * map_tile_size;
	dst.pixels = map->layers[map_layer_midground].cache;
	bak.w = dst.w;
	bak.h = dst.h;
	bak.pixels = map->background;
	gen_background(map, &bak);
	for (i = 0; i < map_layer_count; i++) {
		cache_map_layer(map, &map->layers[i]);
	}
	for (i = 0; i < map->corpse_count; i++) {
		corpse = &map->corpses[i];
		sprite = get_sprite(corpse->sprite);
		src = get_bitmap(sprite->bitmap);
		rcopy_ac(
			&dst,
			src, 
			(int)corpse->x,
			(int)corpse->y,
			&sprite->rect,
			0
		);
	}
}

void init_map_layer(Map* map, Map_Layer* layer) {
	int e, i, z = 0;
	e = map->w * map->h;
	for (i = 0; i < e; i++)
		layer->tiles[i] = map_tile_air;
	e = map->w * map->h * map_tile_size * map_tile_size;
	for (i = 0; i < e; i++)
		layer->cache[i] = *(Colour*)&z;
}

void render_map_background(Map* map) {
	int w, h;
	Bitmap b;
	Rectangle rect;
	profiler_begin("render_map_background");
	w = map->w;
	h = map->h;
	b.pixels = map->background;
	b.w = w * map_tile_size;
	b.h = h * map_tile_size;
	rect.x = 0;
	rect.y = 0;
	rect.w = b.w;
	rect.h = b.h;
	rc_add_cmd_bmp_cp(
		&b,
		0,
		0,
		&rect
	);
	profiler_end();
}

void render_map_blood(Map* map) {
	int w, h;
	Bitmap b;
	Rectangle rect;
	profiler_begin("render_map_blood");
	w = map->w;
	h = map->h;
	b.pixels = map->blood;
	b.w = w * map_tile_size;
	b.h = h * map_tile_size;
	rect.x = 0;
	rect.y = 0;
	rect.w = b.w;
	rect.h = b.h;
	rc_add_cmd_bmp_cp(
		&b,
		0,
		0,
		&rect
	);
	profiler_end();
}

void render_map_layer(Map* map, Map_Layer* layer) {
	int w, h;
	Bitmap b;
	Rectangle rect;
	profiler_begin("render_map_layer");
	w = map->w;
	h = map->h;
	b.pixels = layer->cache;
	b.w = w * map_tile_size;
	b.h = h * map_tile_size;
	rect.x = 0;
	rect.y = 0;
	rect.w = b.w;
	rect.h = b.h;
	rc_add_cmd_bmp_cp(
		&b,
		0,
		0,
		&rect
	);
	profiler_end();
}

void cache_map_layer(Map* map, Map_Layer* layer) {
	int x, y, w, h, t, z = 0;
	Bitmap dst;
	Rectangle rect;
	const Tile_Offset* off;
	const Bitmap* src;
	profiler_begin("cache_map_layer");
	w = map->w;
	h = map->h;
	y = w * h * map_tile_size * map_tile_size;
	for (x = 0; x < y; x++)
		layer->cache[x] = *(Colour*)&z;
	dst.w = w * map_tile_size;
	dst.h = h * map_tile_size;
	dst.pixels = layer->cache;
	src = get_bitmap(asset_id_map);
	rect.w = map_tile_size;
	rect.h = map_tile_size;
	for (y = 0; y < h; y++) {
		for (x = 0; x < w; x++) {
			t = layer->tiles[x + y * w];
			if (!t) continue;
			off = &tile_offsets[t];
			rect.x = off->x;
			rect.y = off->y;
			rcopy_ac(
				&dst,
				src,
				x * map_tile_size,
				y * map_tile_size,
				&rect,
				0
			);
		}
	}
	profiler_end();
}

void map_add_corpse(
	Map* map,
	int x,
	int y,
	Sprite_ID sprite
) {
	Corpse* corpse;
#ifdef DEBUG
	if (map->corpse_count >= map_max_corpses) {
		print_err("Too much death.\n");
		pbreak(error_gameplay_error);
	}
#endif
	corpse = &map->corpses[map->corpse_count++];
	corpse->x = (unsigned short)x;
	corpse->y = (unsigned short)y;
	corpse->sprite = sprite;
}

void map_add_blood(
	Map* map,
	int x,
	int y,
	Asset_ID bitmap,
	const Rectangle* rect
) {
	Bitmap dst;
	dst.w = map->w * map_tile_size;
	dst.h = map->h * map_tile_size;
	dst.pixels = map->blood;
	rcopy_ac(
		&dst,
		get_bitmap(bitmap),
		x,
		y,
		rect,
		0
	);
}
