#include <string.h>

#include "animation.h"
#include "boss.h"
#include "cfgparse.h"
#include "mapobject.h"
#include "maths.h"
#include "sprite.h"

static const char* mo_names[] = {
#define x(type, factory, editor, spawner, params) \
	#type,
map_objects_xmacro()
#undef x
};

static const Map_Object_Factory mo_factories[] = {
#define x(type, factory, editor, spawner, params) \
	factory,
map_objects_xmacro()
#undef x
};

static const int mo_sizes[] = {
#define x(type, factory, editor, spawner, params) \
	sizeof(type ## _paramstruct),
map_objects_xmacro()
#undef x
};

Map_Object_Type get_map_object_type(const char* name) {
	int i;
	for (i = 0; i < map_object_type_count; i++) {
		if (!strcmp(mo_names[i], name)) {
			return (Map_Object_Type)i;
		}
	}
	return map_object_type_count;
}

const char* get_map_object_name(Map_Object_Type type) {
	return mo_names[type];
}

int get_map_object_size(Map_Object_Type type) {
	return sizeof(Map_Object) + mo_sizes[type];
}

Map_Object* create_map_object(
	const char* name,
	const cfg_Object* obj,
	Arena* arena
) {
	Map_Object_Factory f;
	Map_Object_Type type;
	Map_Object* mo;
	type = get_map_object_type(name);
	f = mo_factories[type];
	mo = f(obj, arena);
	mo->type = type;
	return mo;
}

Map_Object* create_prop(
	const cfg_Object* obj,
	Arena* arena
) {
	Map_Object* mo;
	const cfg_Prop* ap;
	prop_paramstruct* params;
	mo = arena_alloc(arena, sizeof *mo + sizeof *params);
	params = (void*)(&mo[1]);
	params->x = find_int_default(obj, "x", 0);
	params->y = find_int_default(obj, "y", 0);
	params->collidable = find_int_default(
		obj,
		"collidable",
		0
	);
	params->destructable = find_int_default(
		obj,
		"destructable",
		0
	);
	params->random = find_int_default(
		obj,
		"randomise_start",
		0
	);
	ap = find_prop_of(
		obj,
		"animation",
		cfg_type_string
	);
	if (ap) {
		params->sprite = get_animation_id(ap->as.str);
		params->anim = 1;
	} else
		params->sprite =
			get_sprite_id(
				find_string_default(obj, "sprite", "sprite_game_over")
			);
	return mo;
}

Map_Object* create_cop(
	const cfg_Object* obj,
	Arena* arena
) {
	Map_Object* mo;
	cop_paramstruct* params;
	mo = arena_alloc(arena, sizeof *mo + sizeof *params);
	params = (void*)(&mo[1]);
	params->x = find_int_default(obj, "x", 0);
	params->y = find_int_default(obj, "y", 0);
	return mo;
}

Map_Object* create_killzone(
	const cfg_Object* obj,
	Arena* arena
) {
	Map_Object* mo;
	killzone_paramstruct* params;
	mo = arena_alloc(arena, sizeof *mo + sizeof *params);
	params = (void*)(&mo[1]);
	params->sprite =
		get_sprite_id(
			find_string_default(obj, "sprite", "sprite_spike_a")
		);
	params->x  = find_int_default(obj, "x", 0);
	params->y  = find_int_default(obj, "y", 0);
	params->ox = find_int_default(obj, "ox", 0);
	params->oy = find_int_default(obj, "oy", 0);
	params->w  = find_int_default(obj, "w", 16 << fbits);
	params->h  = find_int_default(obj, "h", 16 << fbits);
	return mo;
}

Map_Object* create_conttrig(
	const cfg_Object* obj,
	Arena* arena
) {
	Map_Object* mo;
	conttrig_paramstruct* params;
	mo = arena_alloc(arena, sizeof *mo + sizeof *params);
	params = (void*)(&mo[1]);
	params->x = find_int_default(obj, "x", 0);
	params->y = find_int_default(obj, "y", 0);
	params->w = find_int_default(obj, "w", 16 << fbits);
	params->h = find_int_default(obj, "h", 16 << fbits);
	strcpy(
		params->level,
		find_string_default(obj, "level", "level1.lvl")
	);
	return mo;
}

Map_Object* create_gamesaver(
	const cfg_Object* obj,
	Arena* arena
) {
	Map_Object* mo;
	gamesaver_paramstruct* params;
	mo = arena_alloc(arena, sizeof *mo + sizeof *params);
	params = (void*)(&mo[1]);
	params->x = find_int_default(obj, "x", 0);
	params->y = find_int_default(obj, "y", 0);
	params->w = find_int_default(obj, "w", 16 << fbits);
	params->h = find_int_default(obj, "h", 16 << fbits);
	return mo;
}

Map_Object* create_pickup(
	const cfg_Object* obj,
	Arena* arena
) {
	Map_Object* mo;
	pickup_paramstruct* params;
	mo = arena_alloc(arena, sizeof *mo + sizeof *params);
	params = (void*)(&mo[1]);
	params->x = find_int_default(obj, "x", 0);
	params->y = find_int_default(obj, "y", 0);
	params->item = find_int_default(obj, "item", 0);
	return mo;
}

Map_Object* create_cuttrig(
	const cfg_Object* obj,
	Arena* arena
) {
	Map_Object* mo;
	cuttrig_paramstruct* params;
	mo = arena_alloc(arena, sizeof *mo + sizeof *params);
	params = (void*)(&mo[1]);
	params->x = find_int_default(obj, "x", 0);
	params->y = find_int_default(obj, "y", 0);
	params->w = find_int_default(obj, "w", 16 << fbits);
	params->h = find_int_default(obj, "h", 16 << fbits);
	strcpy(
		params->cut,
		find_string_default(obj, "scene", "cut_intro.cut")
	);
	return mo;
}

Map_Object* create_boss(
	const cfg_Object* obj,
	Arena* arena
) {
	Map_Object* mo;
	boss_paramstruct* params;
	mo = arena_alloc(arena, sizeof *mo + sizeof *params);
	params = (void*)(&mo[1]);
	params->x = find_int_default(obj, "x", 0);
	params->y = find_int_default(obj, "y", 0);
	params->type = get_boss_type(
		find_string_default(obj, "boss_type", "boss_type_cia")
	);
	return mo;
}
