#include "cfg.h"
#include "config.h"
#include "maths.h"
#include "plat.h"
#include "rcache.h"
#include "render.h"
#include "sound.h"

extern void prog_init(void*);
extern void prog_update();
extern void prog_deinit();

#if defined(plat_posix)
#define _POSIX_SOURCE
#define _GNU_SOURCE

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>

static clockid_t global_clock;
static unsigned long global_freq;

static void init_timer() {
	struct timespec ts;

	global_clock = CLOCK_REALTIME;
	global_freq = 1000000000;

#if defined(_POSIX_MONOTONIC_CLOCK)
	if (clock_gettime(CLOCK_MONOTONIC, &ts) == 0) {
		global_clock = CLOCK_MONOTONIC;
	}
#endif
}

static unsigned long get_timer() {
	struct timespec ts;

	clock_gettime(global_clock, &ts);
	return
		(unsigned long)ts.tv_sec * global_freq +
		(unsigned long)ts.tv_nsec;
}

unsigned long get_current_time() {
	return get_timer();
}

void sleep_ns(unsigned long ns) {
	struct timespec t = { 0 };
	t.tv_nsec = ns;
	nanosleep(&t, &t);
}

pthread_t audio_thread;
pthread_mutex_t audio_mutex;

void int_plat_start_audio(Int_Audio_Worker w, void* p) {
	int r = pthread_create(
		&audio_thread,
		0,
		w,
		p
	);
	if (r != 0) {
		print_err("Error creating audio thread.");
		pbreak(error_multithreading);
		return;
	}
	pthread_mutex_init(&audio_mutex, 0);
}

void int_plat_wait_audio(void) {
	pthread_mutex_destroy(&audio_mutex);
	pthread_join(audio_thread, 0);
}

void int_plat_lock_audio(void) {
	pthread_mutex_lock(&audio_mutex);
}

void int_plat_unlock_audio(void) {
	pthread_mutex_unlock(&audio_mutex);
}

#include "std_printers.c"

struct File {
	int handle;
	int cursor;
	int size;
};

File* file_open(const char* name, File_Flags flags) {
	File* file;
	int handle, f = 0;
	struct stat info;

	if (flags & file_flags_write) {
		f |= O_WRONLY | O_TRUNC;
	}

	if (flags & file_flags_read) {
		f |= O_RDONLY;
	}

	if (flags & file_flags_read_write) {
		f |= O_RDWR;
	}

	if (stat(name, &info) < 0) {
		info.st_size = 0;
		if (flags & file_flags_create) {
			f |= O_CREAT | S_IRUSR | S_IWUSR;
		}
	}

	handle = open(name, f);
	if (handle < 0) {
		return 0;
	}

	file = (File*)malloc(sizeof *file);
	file->handle = handle;
	file->cursor = 0;
	file->size = (int)info.st_size;

	return file;
}

void file_close(File* file) {
	close(file->handle);
	free(file);
}

void file_read(File* file, void* buffer, int size) {
	lseek(file->handle, (off_t)file->cursor, SEEK_SET);
	read(file->handle, buffer, size);
	file->cursor += size;
}

void file_write(File* file, const void* buffer, int size) {
	lseek(file->handle, (off_t)file->cursor, SEEK_SET);
	write(file->handle, buffer, size);
	file->cursor += size;
}

void file_seek(File* file, int offset, File_Whence whence) {
	switch (whence) {
		case file_whence_begin:
			file->cursor = offset;
			break;
		case file_whence_end:
			file->cursor = file->size - offset;
			break;
		case file_whence_cur:
			file->cursor += offset;
			break;
	}
}

int file_size(File* file) {
	return file->size;
}

int file_exists(const char* name) {
	struct stat info;
	return stat(name, &info) == 0;
}

#endif

#if defined(plat_x11)
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define txt_buf_max 32

struct {
	Display* display;
	Window window;
	GC gc;
	Atom wm_protocols, wm_delete;
	int w, h, running, fps;
	int mx, my, scrx, scry;
	XImage* bb;
	Colour* pixels;
	char txt_buf[txt_buf_max];
	char raw_buf[txt_buf_max];
	int txt_len;
#if debug_redraw
	XColor redraw_col;
	XRectangle redraw_rects[max_debug_redraw_rects];
	int redraw_count;
#endif
} windowing;

static void init_rendering(XWindowAttributes* wa) {
	Colour* p;
	Display* d;
	int w, h;

	d = windowing.display;

	w = wa->width;
	h = wa->height;

	p = (Colour*)malloc(sizeof *p * w * h);
	if (!p) {
		print_err("Out of memory.\n");
		pbreak(error_out_of_memory);
	}
	windowing.pixels = p;
	windowing.bb = XCreateImage(
		d,
		wa->visual,
		wa->depth,
		ZPixmap,
		0,
		(char*)p,
		w,
		h,
		32,
		w * sizeof *p
	);
	if (!windowing.bb) {
		print_err("Failed to create X11 backbuffer.\n");
		pbreak(error_platform_error);
	}
}

void plat_present(
	int x,
	int y,
	int w,
	int h
) {
	int rat_x,  rat_y;
	int rrat_x, rrat_y;
	int sx, sy;
	int ix, iy, sw, sh, dw, dh;
	int dsx, dsy;
	int dex, dey;
	int i;
	Colour* dst;
	const Colour* src;
	unsigned char t;
	sw = renderer_w;
	sh = renderer_h;
	dw = windowing.w;
	dh = windowing.h;
	rat_x  = (sw << fbits * 2) / (dw << fbits);
	rat_y  = (sh << fbits * 2) / (dh << fbits);
	rrat_x = (dw << fbits * 2) / (sw << fbits);
	rrat_y = (dh << fbits * 2) / (sh << fbits);
	dsx = (x       * rrat_x) >> fbits;
	dsy = (y       * rrat_y) >> fbits;
	dex = ((x + w) * rrat_x) >> fbits;
	dey = ((y + h) * rrat_y) >> fbits;
	dst = windowing.pixels;
	src = get_render_pixels();
	for (iy = dsy; iy < dey; iy++) {
		sy = (rat_y * (iy << fbits)) >> (fbits * 2);
		for (ix = dsx; ix < dex; ix++) {
			sx = (rat_x * (ix << fbits)) >> (fbits * 2);
			i = sx + sy * sw;
			dst = &windowing.pixels[ix + iy * dw];
			*dst = src[i];
			t = dst->r;
			dst->r = dst->b;
			dst->b = t;
		}
	}
	XPutImage(
		windowing.display,
		windowing.window,
		windowing.gc,
		windowing.bb,
		dsx, dsy,
		dsx, dsy,
		dex - dsx, dey - dsy
	);
}

static Btn btn_from_xkey(unsigned key) {
	switch (key) {
		case XK_Left:
		case 0x68:
			return btn_left;
		case XK_Right:
		case 0x6c:
			return btn_right;
		case XK_Up:
		case 0x6b:
			return btn_up;
		case XK_Down:
		case 0x6a:
			return btn_down;
		case 0x78:
			return btn_shoot;
		case 0x7a:
			return btn_jump;
		case 0x63:
			return btn_special;
		case 0x61:
			return btn_inv;
		case XK_Escape:
			return btn_pause;
		case XK_Shift_L:
		case XK_Shift_R:
			return btn_pc_shift;
		case XK_Control_L:
		case XK_Control_R:
			return btn_pc_ctrl;
		case XK_Alt_L:
		case XK_Alt_R:
			return btn_pc_alt;
		case XK_BackSpace:
			return btn_pc_backspace;
		case XK_Return:
			return btn_pc_return;
		case XK_F12:
			return btn_pc_f12;
		default:
			return btn_unknown;
	}
}

static Btn  held_btns[btn_count];
static Btn  pressed_btns[btn_count];
static Btn  released_btns[btn_count];
static MBtn mheld_btns[mbtn_count];
static MBtn mpressed_btns[mbtn_count];
static MBtn mreleased_btns[mbtn_count];

static void reset_keys() {
	int i;
	for (i = 0; i < btn_count; i++) {
		pressed_btns[i] = 0;
		released_btns[i] = 0;
	}
	for (i = 0; i < mbtn_count; i++) {
		mpressed_btns[i] = 0;
		mreleased_btns[i] = 0;
	}
}

static void handle_events() {
	XEvent e;
	XWindowAttributes wa;
	Rectangle rr = { 0, 0, renderer_w, renderer_h };
	Display* d;
	Window w;
	KeySym sym, lsym;
	int nw, nh, mx, my, rx, ry, sw, sh, dw, dh, ltype, tl;
	int mb;
	Btn btn;
	char* raw, * txt;
	const char* c;

	d = windowing.display;
	w = windowing.window;
	txt = windowing.txt_buf;
	raw = windowing.raw_buf;
	tl = windowing.txt_len;

	sw = renderer_w;
	sh = renderer_h;
	dw = windowing.w;
	dh = windowing.h;

	reset_keys();
	windowing.scrx = windowing.scry = 0;

	ltype = -1;

	while (XPending(d)) {
		XNextEvent(d, &e);

		switch (e.type) {
			case ClientMessage:
				if (
					(Atom)e.xclient.data.l[0] ==
					windowing.wm_delete
				) {
					windowing.running = 0;
				}
				break;
			case Expose:
				XGetWindowAttributes(d, w, &wa);
				nw = wa.width;
				nh = wa.height;
				if (
					!windowing.bb ||
					nw != windowing.w ||
					nh != windowing.h
				) {
					if (windowing.bb)
						XDestroyImage(windowing.bb);
					init_rendering(&wa);
					windowing.w = nw;
					windowing.h = nh;
				}
				rc_invalidate(&rr);
				break;
			case MotionNotify:
				rx = (sw << fbits * 2) / (dw << fbits);
				ry = (sh << fbits * 2) / (dh << fbits);
				mx = e.xmotion.x;
				my = e.xmotion.y;
				mx = ((mx << fbits) * rx) >> fbits * 2;
				my = ((my << fbits) * ry) >> fbits * 2;
				windowing.mx = mx;
				windowing.my = my;
				break;
			case KeyPress:
				sym = XLookupKeysym(&e.xkey, 0);
				btn = btn_from_xkey(sym);
				held_btns[btn] = 1;
				pressed_btns[btn] = 1;
				if (ltype == KeyRelease && lsym == sym) {
					lsym = sym;
					released_btns[btn] = 0;
					pressed_btns[btn] = 0;
				}
				if (tl >= txt_buf_max)
					break;
				XLookupString(
					&e.xkey,
					raw + tl,
					txt_buf_max - tl,
					0,
					0
				);
				for (c = raw + tl; *c; c++) {
					if (*c >= ' ' && *c <= '~') {
						txt[windowing.txt_len++] = *c;
					}
				}
				break;
			case KeyRelease:
				sym = XLookupKeysym(&e.xkey, 0);
				btn = btn_from_xkey(sym);
				held_btns[btn] = 0;
				released_btns[btn] = 1;
				lsym = sym;
				break;
			case ButtonPress:
				switch (e.xbutton.button) {
					case 1:
					case 2:
					case 3:
						mb = e.xbutton.button - 1;
						mheld_btns[mb] = 1;
						mpressed_btns[mb] = 1;
						break;
				}
				break;
			case ButtonRelease:
				switch (e.xbutton.button) {
					case 1:
					case 2:
					case 3:
						mb = e.xbutton.button - 1;
						mheld_btns[mb] = 0;
						mreleased_btns[mb] = 1;
						break;
					case 4:
						windowing.scry--;
						break;
					case 5:
						windowing.scry++;
						break;
					case 6:
						windowing.scrx--;
						break;
					case 7:
						windowing.scrx++;
						break;
				}
				break;
		}

		ltype = e.type;
	}
}

void plat_flush(void) {
	plat_present(0, 0, renderer_w, renderer_h);
	handle_events();
}

static void hide_pointer() {
	Cursor cursor;
	Pixmap blank;
	char data = 0;
	Window w;
	Display* d;
	XColor col;

	d = windowing.display;
	w = windowing.window;

	blank = XCreateBitmapFromData(d, w, &data, 1, 1);
	cursor = XCreatePixmapCursor(
		d,
		blank,
		blank,
		&col,
		&col,
		0,
		0
	);
	XDefineCursor(d, w, cursor);
}

void read_cfg(Cfg* cfg, const char* filename) {
	FILE* f;
	Arena a;
	char* buf;
	int s;
#define cfg_mem 1024
	init_arena(
		&a,
		malloc(cfg_mem),
		cfg_mem
	);
	default_cfg(cfg);
	f = fopen(filename, "r");
	if (!f) return;
	fseek(f, 0, SEEK_END);
	s = ftell(f);
	rewind(f);
	buf = malloc(s + 1);
	s = fread(buf, 1, s, f);
	buf[s] = 0;
	init_cfg(cfg, &a, buf);
	free(a.buffer);
	free(buf);
#undef cfg_mem
}

void init_audio(void);
void deinit_audio(void);

int main() {
	Window root;
	XWindowAttributes wa;
	unsigned rm, bm;
	Display* d;
	Window w;
	Cfg* cfg;
	GC gc;
	int i;
	unsigned frame = 0;
	char* memory;
	XSizeHints* hints;
	unsigned long now, next, end;
	long ts;

	cfg = get_cfg();
	read_cfg(cfg, cfgfile_name);
	init_timer();
	init_audio();

	windowing.running = 0;

	windowing.display = XOpenDisplay(0);
	if (!windowing.display) {
		print_err("Failed to open X11 display.\n");
		return error_platform_error;
	}

	d = windowing.display;

	windowing.wm_protocols = XInternAtom(
		d,
		"WM_PROTOCOLS",
		0
	);
	windowing.wm_delete = XInternAtom(
		d,
		"WM_DELETE_WINDOW",
		0
	);

	root = DefaultRootWindow(d);
	w = XCreateSimpleWindow(
		d,
		root,
		0,
		0,
		cfg->win_w,
		cfg->win_h,
		0,
		WhitePixel(d, 0),
		BlackPixel(d, 0)
	);
	windowing.window = w;

	if (!cfg->resizable) {
		hints = XAllocSizeHints();
		hints->flags = PMinSize | PMaxSize;
		hints->min_width = default_window_w;
		hints->min_height = default_window_h;
		hints->max_width = default_window_w;
		hints->max_height = default_window_h;
		XSetWMNormalHints(d, w, hints);
		XFree(hints);
	}

	XSetWMProtocols(d, w, &windowing.wm_delete, 1);
	XStoreName(d, w, game_name);

	XSelectInput(
		d,
		w,
		ExposureMask      |
		KeyPressMask      |
		KeyReleaseMask    |
		PointerMotionMask |
		ButtonPressMask   |
		ButtonReleaseMask
	);

	XClearWindow(d, w);
	XMapRaised(d, w);

	gc = XCreateGC(d, w, 0, 0);
	if (!gc) {
		print_err("Failed to create X graphics context.\n");
		pbreak(error_platform_error);
	}

	windowing.gc = gc;

	XGetWindowAttributes(d, w, &wa);
	if (wa.depth != 24 && wa.depth != 32) {
		print_err("Only true colour displays are supported.\n");
		pbreak(error_platform_error);
	}
	windowing.w = wa.width;
	windowing.h = wa.height;
	rm = wa.visual->red_mask  & 0x1000000;
	bm = wa.visual->blue_mask & 0xffffff;
	if ((rm == 0xff && bm == 0xff0000)) {
	    print_war("Detected BGR. Colours will look fucked.\n");
	}

#if hide_mouse
	hide_pointer();
#endif

	init_rendering(&wa);

	memory = (char*)malloc(memory_size);
	if (!memory) {
		print_err("Out of memory.\n");
		pbreak(error_out_of_memory);
	}

	for (i = 0; i < btn_count; i++) {
		held_btns[i] = 0;
	}
	for (i = 0; i < mbtn_count; i++) {
		mheld_btns[i] = 0;
	}

	windowing.txt_len = 0;

	windowing.running = 1;
	windowing.fps = 0;

	prog_init(memory);

	next = get_current_time();
	while (windowing.running) {
		now = get_current_time();
		while (now >= next) {
			handle_events();
			int_plat_lock_audio();
			prog_update();
			int_plat_unlock_audio();
#if debug_redraw
			update_debug_redraw();
#endif
			next += ms_per_frame * 1000000;
			frame++;
		}
		ts = next - now;
		if (ts > 0)
			sleep_ns(ts);
		end = get_current_time();
		if (!(frame % 10))
			windowing.fps = 1000000000 / (end - now);
	}

	prog_deinit();

	deinit_audio();
	free(memory);

	XDestroyWindow(d, w);
	XDestroyImage(windowing.bb);
	XCloseDisplay(d);

	return error_none;
}

int plat_fps(void) {
	return windowing.fps;
}

void plat_quit(void) {
	windowing.running = 0;
}

int get_mouse_x() {
	return windowing.mx;
}

int get_mouse_y() {
	return windowing.my;
}

int btn_pressed(Btn btn) {
	return held_btns[btn];
}

int btn_just_pressed(Btn btn) {
	return pressed_btns[btn];
}

int btn_just_released(Btn btn) {
	return released_btns[btn];
}

int any_btn_just_pressed(void) {
	int b;
	for (b = 0; b < btn_count; b++)
		if (pressed_btns[b])
			return 1;
	return 0;
}

int mbtn_pressed(MBtn btn) {
	return mheld_btns[btn];
}

int mbtn_just_pressed(MBtn btn) {
	return mpressed_btns[btn];
}

int mbtn_just_released(MBtn btn) {
	return mreleased_btns[btn];
}

int get_mscroll_x(void) {
	return windowing.scrx;
}

int get_mscroll_y(void) {
	return windowing.scry;
}

const char* get_text_input(int* len) {
	int tl;
	tl = windowing.txt_len;
	if (!tl) {
		return 0;
	}
	if (len) { *len = tl; }
	windowing.txt_len = 0;
	return windowing.txt_buf;
}

#endif

#ifdef plat_pulse
#define inline /* erm sir we are using C90 here... */
#include <stdlib.h>
#include <pulse/simple.h>

struct {
	unsigned char* buf;
	short* buf2;
	pa_simple* dev;
	int r;
} audio;

void* audio_worker(void* arg) {
	const int s = audio_buffer_size;
	const Cfg* cfg;
	int i, c = 1, v;
	unsigned char* buf = audio.buf;
	short* buf2 = audio.buf2;
	pa_simple* dev = audio.dev;
	cfg = get_cfg();
	v = cfg->audio;
	(void)arg;
	while (c) {
		int_plat_lock_audio();
		if (!audio.r) c = 0;
		sound_mix(buf, s);
		int_plat_unlock_audio();
		for (i = 0; i < s; i++) {
			buf2[i] = (buf[i] * 2 - 256) * 80 * v;
		}
		pa_simple_write(dev, buf2, s * sizeof(short), 0);
	}
	return 0;
}

void init_audio() {
	pa_sample_spec sp = { 0 };
	sp.format = PA_SAMPLE_S16LE;
	sp.channels = 1;
	sp.rate = audio_sample_rate;
	audio.dev = pa_simple_new(
		0,
		game_name,
		PA_STREAM_PLAYBACK,
		0,
		"Game Audio",
		&sp,
		0,
		0,
		0
	);
	audio.r = 0;
	if (!audio.dev) {
		print_err("Failed to create audio device.\n");
		print_war("Running without audio.\n");
		return;
	}
	audio.r = 1;
	audio.buf = malloc(audio_buffer_size);
	audio.buf2 = malloc(
		audio_buffer_size *
		sizeof *audio.buf2
	);
	int_plat_start_audio(audio_worker, &audio);
}

void deinit_audio() {
	if (audio.dev) {
		audio.r = 0;
		int_plat_wait_audio();
		pa_simple_drain(audio.dev, 0);
		pa_simple_free(audio.dev);
		free(audio.buf);
		free(audio.buf2);
	}
}

#endif

#ifdef plat_win
 /* Rectangle struct conflicts with a Win32 function... */
#define Rectangle WinRectangle
#include <windows.h>
#undef Rectangle

#if opengl_enable
#include <GL/gl.h>
#endif

#include <stdlib.h>
#include <stdio.h>

struct File {
	HFILE handle;
	int size;
};

File* file_open(const char* name, File_Flags flags) {
	File* file;
	UINT f = 0;
	DWORD s, r;
	HFILE h;
	OFSTRUCT info;
	if (flags & file_flags_write)
		f |= OF_WRITE | OF_CREATE | OF_SHARE_DENY_WRITE;
	if (flags & file_flags_read)
		f |= OF_READ;
	if (flags & file_flags_read_write)
		f |= OF_READWRITE | OF_SHARE_DENY_WRITE;
	h = OpenFile(name, &info, f);
	if (h == HFILE_ERROR)
		return 0;
	if (GetFileSize(h, &s) == INVALID_FILE_SIZE) {
		CloseHandle(h);
		return 0;
	}
	file = malloc(sizeof *file);
	file->handle = h;
	file->size = (int)s;
	return file;
}

void file_close(File* file) {
	CloseHandle(file->handle);
	free(file);
}

void file_read(File* file, void* buffer, int size) {
	DWORD r;
	ReadFile(file->handle, buffer, size, &r, 0);
}

void file_write(File* file, const void* buffer, int size) {
	DWORD w;
	WriteFile(file->handle, buffer, size, &w, 0);
}

void file_seek(File* file, int offset, File_Whence whence) {
	SetFilePointer(
		file->handle,
		(LONG)offset,
		0,
		whence == file_whence_begin? FILE_BEGIN:
		whence == file_whence_end? FILE_END:
		FILE_CURRENT
	);
}

int file_size(File* file) {
	return file->size;
}

int file_exists(const char* name) {
	return PathFileExistsA(name);
}

#define txt_buf_max 32

struct {
	HINSTANCE i;
	HWND w;
	HBITMAP bmp;
	HDC bdc;
	Colour* pixels;
	int o, txt_len;
	int wi, he;
	int mx, my, scrx, scry;
	int is_init;
	int fps;
	DWORD style, ex_style;
	HANDLE audio_thread, audio_mutex, audio_sem;
	HWAVEOUT audio_dev;
	WAVEHDR audio_bufs[2];
	int audio_buf_count, audio_buf, audio_run;
	BYTE kstate[256];
	char txt_buf[txt_buf_max];
	char raw_buf[txt_buf_max];
#if opengl_enable
	HGLRC rc;
#endif
} app;

static Btn  held_btns[btn_count];
static Btn  pressed_btns[btn_count];
static Btn  released_btns[btn_count];
static MBtn mheld_btns[mbtn_count];
static MBtn mpressed_btns[mbtn_count];
static MBtn mreleased_btns[mbtn_count];

static void reset_keys() {
	int i;
	for (i = 0; i < btn_count; i++) {
		pressed_btns[i] = 0;
		released_btns[i] = 0;
	}
	for (i = 0; i < mbtn_count; i++) {
		mpressed_btns[i] = 0;
		mreleased_btns[i] = 0;
	}
}

static Btn btn_from_vk(unsigned k) {
	switch (k) {
		case VK_LEFT:
		case 0x48:
			return btn_left;
		case VK_RIGHT:
		case 0x4c:
			return btn_right;
		case VK_UP:
		case 0x4b:
			return btn_up;
		case VK_DOWN:
		case 0x4a:
			return btn_down;
		case 0x5a:
			return btn_jump;
		case 0x58:
			return btn_shoot;
		case 0x43:
			return btn_special;
		case 0x41:
			return btn_inv;
		case VK_ESCAPE:
			return btn_pause;
		case VK_LSHIFT:
		case VK_RSHIFT:
			return btn_pc_shift;
		case VK_LCONTROL:
		case VK_RCONTROL:
			return btn_pc_ctrl;
		case VK_LMENU:
		case VK_RMENU:
			return btn_pc_alt;
		case VK_BACK:
			return btn_pc_backspace;
		case VK_RETURN:
			return btn_pc_return;
		case VK_F12:
			return btn_pc_f12;
		default:
			return btn_unknown;
	}
}

static void init_bmp(void) {
	BITMAPINFO h = { 0 };
	HDC dc;
	if (app.bmp)
		DeleteObject(app.bmp);
	h.bmiHeader.biSize = sizeof h;
	h.bmiHeader.biWidth = app.wi;
	h.bmiHeader.biHeight = -app.he;
	h.bmiHeader.biPlanes = 1;
	h.bmiHeader.biSizeImage = app.wi * app.he;
	h.bmiHeader.biBitCount =
		sizeof *app.pixels * 8;
	dc = GetDC(app.w);
	app.bmp = CreateDIBSection(
		dc,
		&h,
		DIB_RGB_COLORS,
		(void**)(&app.pixels),
		0,
		0
	);
	app.bdc = CreateCompatibleDC(dc);
	SelectObject(app.bdc, app.bmp);
}

static LRESULT evt_cbk(
	HWND w,
	UINT msg,
	WPARAM wparam,
	LPARAM lparam
) {
	HDC dc;
	PAINTSTRUCT paint;
	RECT r;
	Btn b;
	Rectangle rr = { 0, 0, renderer_w, renderer_h };
	int rx, ry, sw, sh, dw, dh, mx, my, i, tl, wi, he;
	char* raw, * txt;
	const char* c;
	MINMAXINFO* mmi;
	sw = renderer_w;
	sh = renderer_h;
	dw = app.wi;
	dh = app.he;
	txt = app.txt_buf;
	raw = app.raw_buf;
	tl =  app.txt_len;
	switch (msg) {
		case WM_CLOSE:
			app.o = 0;
			return 0;
#if !opengl_enable
		case WM_PAINT:
			dc = BeginPaint(w, &paint);
			r = paint.rcPaint;
			wi = r.right  - r.left;
			he = r.bottom - r.top;
			BitBlt(
				dc,
				r.left,
				r.top,
				wi,
				he,
				app.bdc,
				r.left,
				r.top,
				SRCCOPY
			);
			EndPaint(w, &paint);
			return 0;
#endif
		case WM_SIZE:
			app.wi = lparam & 0xFFFF;
			app.he = (lparam >> 16) & 0xFFFF;
#if opengl_enable
			glViewport(0, 0, app.wi, app.he);
#else
			init_bmp();
			if (app.is_init)
				rc_invalidate(&rr);
#endif
			return 0;
		case WM_GETMINMAXINFO:
			mmi = (MINMAXINFO*)lparam;
			mmi->ptMinTrackSize.x = renderer_w;
			mmi->ptMinTrackSize.y = renderer_h;
			return 0;
		case WM_KEYDOWN:
			b = btn_from_vk((unsigned)wparam);
			held_btns[b] = 1;
			pressed_btns[b] = 1;
			GetKeyboardState(app.kstate);
			/* this is unsafe. too bad. */
			i = ToAscii(
				(unsigned)wparam,
				(lparam >> 16) & 0xff,
				app.kstate,
				(LPWORD)(raw + tl),
				0
			);
			raw[i] = '\0';
			for (c = raw + tl; *c; c++) {
				if (*c >= ' ' && *c <= '~') {
					txt[app.txt_len++] = *c;
				}
			}
			return 0;
		case WM_KEYUP:
			b = btn_from_vk((unsigned)wparam);
			held_btns[b] = 0;
			released_btns[b] = 1;
			return 0;
		case WM_MOUSEMOVE:
			rx = (sw << fbits * 2) / (dw << fbits);
			ry = (sh << fbits * 2) / (dh << fbits);
			mx = (int)(lparam & 0xffff);
			my = ((lparam >> 16) & 0xffff);
			mx = ((mx << fbits) * rx) >> fbits * 2;
			my = ((my << fbits) * ry) >> fbits * 2;
			app.mx = mx;
			app.my = my;
			return 0;
		case WM_MOUSEWHEEL:
			app.scry += GET_WHEEL_DELTA_WPARAM(wparam) > 1 ? -1 : 1;
			return 0;
		case WM_LBUTTONDOWN:
			mheld_btns[mbtn_left] = 1;
			mpressed_btns[mbtn_left] = 1;
			return 0;
		case WM_MBUTTONDOWN:
			mheld_btns[mbtn_middle] = 1;
			mpressed_btns[mbtn_middle] = 1;
			return 0;
		case WM_RBUTTONDOWN:
			mheld_btns[mbtn_right] = 1;
			mpressed_btns[mbtn_right] = 1;
			return 0;
		case WM_LBUTTONUP:
			mheld_btns[mbtn_left] = 0;
			mreleased_btns[mbtn_left] = 1;
			return 0;
		case WM_MBUTTONUP:
			mheld_btns[mbtn_middle] = 0;
			mreleased_btns[mbtn_middle] = 1;
			return 0;
		case WM_RBUTTONUP:
			mheld_btns[mbtn_right] = 0;
			mreleased_btns[mbtn_right] = 1;
			return 0;
		default:
			return DefWindowProc(w, msg, wparam, lparam);
	}
}

#if opengl_enable
int init_opengl(void) {
	PIXELFORMATDESCRIPTOR pfd = { 0 };
	HDC dc;
	int fmt, r;
	pfd.nSize = sizeof pfd;
	pfd.nVersion = 1;
	pfd.dwFlags =
		PFD_DRAW_TO_WINDOW |
		PFD_SUPPORT_OPENGL |
		PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 24;
	dc = GetDC(app.w);
	fmt = ChoosePixelFormat(dc, &pfd);
	if (!fmt) return 0;
	r = DescribePixelFormat(dc, fmt, sizeof pfd, &pfd);
	if (!r) return 0;
	r = SetPixelFormat(dc, fmt, &pfd);
	if (!r) return 0;
	app.rc = wglCreateContext(dc);
	if (!app.rc) return 0;
	wglMakeCurrent(dc, app.rc);
	app.bdc = dc;
	return 1;
}

void deinit_opengl(void) {
	wglDeleteContext(app.rc);
	ReleaseDC(app.w, app.bdc);
}
#endif

static void init_app(
	Cfg* cfg,
	HINSTANCE i,
	void* mem
) {
	WNDCLASSA wc = { 0 };
	DWORD
	style =
		WS_CAPTION |
		WS_SYSMENU |
		WS_MINIMIZEBOX |
		WS_VISIBLE,
	ex_style =
		WS_EX_APPWINDOW |
		WS_EX_WINDOWEDGE;
	RECT rect = { 0 };
	int w, h;
	if (cfg->resizable)
		style |=
			WS_MAXIMIZEBOX |
			WS_THICKFRAME;
	rect.right  = cfg->win_w;
	rect.bottom = cfg->win_h;
	app.i = i;
	app.o = 0;
	wc.hIcon = LoadIcon(0, IDI_APPLICATION);
	wc.hCursor = LoadCursor(0, IDC_ARROW);
	wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.hInstance = i;
	wc.lpfnWndProc = evt_cbk;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.lpszMenuName = 0;
	wc.hbrBackground = 0;
	wc.lpszClassName = game_name;
	RegisterClassA(&wc);
	AdjustWindowRectEx(&rect, style, 0, ex_style);
	w = rect.right  - rect.left;
	h = rect.bottom - rect.top;
	app.style = style;
	app.ex_style = ex_style;
	app.w = CreateWindowExA(
		ex_style,
		game_name,
		game_name,
		style,
		300,
		300,
		w,
		h,
		0,
		0,
		i,
		0
	);
	app.wi = cfg->win_w;
	app.he = cfg->win_h;
	ShowWindow(app.w, SW_SHOWNORMAL);
	app.bmp = 0;
	init_bmp();
	app.o = 1;
	app.mx = 0;
	app.my = 0;
	app.fps = 0;
#if opengl_enable
	if (!init_opengl()) {
		print_err("Failed to init OpenGL.\n");
		pbreak(error_platform_error);
	}
#endif
}

static void deinit_app(void) {
#if opengl_enable
	deinit_opengl();
#endif
	DeleteObject(app.bmp);
	DestroyWindow(app.w);
	UnregisterClass(game_name, app.i);
}

void plat_present(
	int x,
	int y,
	int w,
	int h
) {
#if opengl_enable
	(void)x; (void)y; (void)w; (void)h;
#else
	int rat_x,  rat_y;
	int rrat_x, rrat_y;
	int sx, sy;
	int ix, iy, sw, sh, dw, dh;
	int dsx, dsy;
	int dex, dey;
	int i;
	Colour* dst;
	const Colour* src;
	unsigned char t;
	RECT r;
	sw = renderer_w;
	sh = renderer_h;
	dw = app.wi;
	dh = app.he;
	rat_x  = (sw << fbits * 2) / (dw << fbits);
	rat_y  = (sh << fbits * 2) / (dh << fbits);
	rrat_x = (dw << fbits * 2) / (sw << fbits);
	rrat_y = (dh << fbits * 2) / (sh << fbits);
	dsx = (x       * rrat_x) >> fbits;
	dsy = (y       * rrat_y) >> fbits;
	dex = ((x + w) * rrat_x) >> fbits;
	dey = ((y + h) * rrat_y) >> fbits;
	dst = app.pixels;
	src = get_render_pixels();
	for (iy = dsy; iy < dey; iy++) {
		sy = (rat_y * (iy << fbits)) >> (fbits * 2);
		for (ix = dsx; ix < dex; ix++) {
			sx = (rat_x * (ix << fbits)) >> (fbits * 2);
			i = sx + sy * sw;
			dst = &app.pixels[ix + iy * dw];
			*dst = src[i];
			t = dst->r;
			dst->r = dst->b;
			dst->b = t;
		}
	}
	r.left   = dsx;
	r.top    = dsy;
	r.right  = dex;
	r.bottom = dey;
	InvalidateRect(app.w, &r, FALSE);
#endif
}

static void update_app(void) {
	MSG msg;
	reset_keys();
	app.scrx = app.scry = 0;
	while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
}

void plat_flush(void) {
	update_app();
}

unsigned long long get_ticks() {
	LARGE_INTEGER ticks;
	QueryPerformanceCounter(&ticks);
	return (unsigned long long)ticks.QuadPart;
}

unsigned long long get_freq() {
	LARGE_INTEGER f;
	QueryPerformanceFrequency(&f);
	return (unsigned long long)f.QuadPart;
}

void read_cfg(Cfg* cfg, const char* filename) {
	FILE* f;
	Arena a;
	char* buf;
	int s;
#define cfg_mem 1024
	init_arena(
		&a,
		malloc(cfg_mem),
		cfg_mem
	);
	default_cfg(cfg);
	f = fopen(filename, "r");
	if (!f) return;
	fseek(f, 0, SEEK_END);
	s = ftell(f);
	rewind(f);
	buf = malloc(s + 1);
	s = fread(buf, 1, s, f);
	buf[s] = 0;
	init_cfg(cfg, &a, buf);
	free(a.buffer);
	free(buf);
#undef cfg_mem
}

void audio_cbk(
	HWAVEOUT wo,
	UINT msg,
	DWORD_PTR inst,
	DWORD_PTR param1,
	DWORD_PTR param2
) {
	if (msg != WOM_DONE) return;
	ReleaseSemaphore(app.audio_sem, 1, 0);
}

int init_audio(void) {
	WAVEFORMATEX fmt;
	WAVEHDR* hdr;
	int s = 1, i;
	fmt.wFormatTag = WAVE_FORMAT_PCM;
	fmt.nChannels = 1;
	fmt.nSamplesPerSec = audio_sample_rate;
	fmt.wBitsPerSample = 8;
	fmt.nBlockAlign =
		fmt.nChannels * fmt.wBitsPerSample / 8;
	fmt.nAvgBytesPerSec =
		fmt.nBlockAlign * fmt.nSamplesPerSec;
	fmt.cbSize = 0;
	waveOutOpen(
		&app.audio_dev,
		WAVE_MAPPER,
		&fmt,
		(DWORD_PTR)audio_cbk,
		(DWORD_PTR)app.i,
		CALLBACK_FUNCTION
	);
	if (!app.audio_dev) return 0;
	for (i = 0; i < app.audio_buf_count; i++) {
		hdr = &app.audio_bufs[i];
		memset(hdr, 0, sizeof *hdr);
		hdr->lpData = malloc(audio_buffer_size); 
		hdr->dwBufferLength = audio_buffer_size;
		waveOutPrepareHeader(
			app.audio_dev,
			hdr,
			sizeof *hdr
		);
		s = s && hdr->lpData;
	}
	app.audio_buf = 0;
	return s;
}

void deinit_audio(void) {
	int i;
	WAVEHDR* hdr;
	if (!app.audio_run) return;
	waveOutClose(app.audio_dev);
	for (i = 0; i < app.audio_buf_count; i++) {
		hdr = &app.audio_bufs[i];
		free(hdr->lpData);
		waveOutUnprepareHeader(
			app.audio_dev,
			hdr,
			sizeof *hdr
		);
	}
}

void* audio_worker(void* ptr) {
	int r = 1;
	WAVEHDR* hdr;
	(void)ptr;
	app.audio_run = 0;
	if (!init_audio()) return 0;
	app.audio_run = 1;
	while (r) {
		hdr = &app.audio_bufs[app.audio_buf];
		int_plat_lock_audio();
		r = app.o;
		sound_mix(
			hdr->lpData,
			audio_buffer_size
		);
		int_plat_unlock_audio();
		waveOutWrite(
			app.audio_dev,
			hdr,
			sizeof * hdr
		);
		app.audio_buf =
			(app.audio_buf + 1) % app.audio_buf_count;
		WaitForSingleObject(app.audio_sem, INFINITE);
	}
	deinit_audio();
	return 0;
}

void int_plat_start_audio(Int_Audio_Worker w, void* p) {
	app.audio_buf_count = 2;
	app.audio_mutex = CreateMutex(
		0,
		0,
		"Audio Mutex"
	);
	app.audio_sem = CreateSemaphore(
		0,
		app.audio_buf_count - 1,
		app.audio_buf_count,
		NULL
	);
	app.audio_thread = CreateThread(
		0,
		0,
		(LPTHREAD_START_ROUTINE)w,
		p,
		0,
		0
	);
}

void int_plat_wait_audio(void) {
	WaitForSingleObject(app.audio_thread, INFINITE);
	WaitForSingleObject(app.audio_mutex, INFINITE);
}

void int_plat_lock_audio(void) {
	WaitForSingleObject(app.audio_mutex, INFINITE);
}

void int_plat_unlock_audio(void) {
	ReleaseMutex(app.audio_mutex);
}

int WinMain(
	HINSTANCE instance,
	HINSTANCE prev_instance,
	LPSTR cmd,
	int show_cmd
) {
	int i;
	void* mem;
	unsigned long long now, next, end, f;
	long long ns, t;
	unsigned frame;
#if opengl_enable
	unsigned tex;
#endif
	Cfg cfg;
	HANDLE timer;
	LARGE_INTEGER dt;
	read_cfg(&cfg, cfgfile_name);
	f = get_freq();
	mem = malloc(memory_size);
	if (!mem) {
		print_err("Out of memory.\n");
		pbreak(error_out_of_memory);
	}
	for (i = 0; i < btn_count; i++)
		held_btns[i] = 0;
	for (i = 0; i < mbtn_count; i++)
		mheld_btns[i] = 0;
	app.is_init = 0;
	init_app(&cfg, instance, mem);
	int_plat_start_audio(audio_worker, 0);
	prog_init(mem);
	app.is_init = 1;
	timer = CreateWaitableTimer(0, TRUE, 0);
#if opengl_enable
	glEnable(GL_TEXTURE_2D);
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
#endif
	timeBeginPeriod(1);
	next = get_ticks() * 1000000000 / f;
	while (app.o) {
		now = get_ticks() * 1000000000 / f;
		while (now >= next) {
			int_plat_lock_audio();
			update_app();
			prog_update();
			int_plat_unlock_audio();
#if opengl_enable
			glTexImage2D(
				GL_TEXTURE_2D,
				0,
				GL_RGBA,
				renderer_w,
				renderer_h,
				0,
				GL_RGBA,
				GL_UNSIGNED_BYTE,
				get_render_pixels()
			);
			glColor3f(1.0f, 1.0f, 1.0f);
			glBegin(GL_QUADS);
				glTexCoord2f(0.0f, 1.0f);
				glVertex2f(-1.0f, -1.0f);
				glTexCoord2f(1.0f, 1.0f);
				glVertex2f(1.0f, -1.0f);
				glTexCoord2f(1.0f, 0.0f);
				glVertex2f(1.0f, 1.0f);
				glTexCoord2f(0.0f, 0.0f);
				glVertex2f(-1.0f, 1.0f);
			glEnd();
			SwapBuffers(app.bdc);
#endif
			next += ms_per_frame * 1000000;
			frame++;
		}
		t = maxi(0, (next - now)) / 100;
		if (t > 0) {
			dt.QuadPart = -t;
			timeBeginPeriod(1);
			if (SetWaitableTimerEx(timer, &dt, 0, 0, 0, 0, 0)) {
				WaitForSingleObject(timer, INFINITE);
			}
			timeEndPeriod(1);
		}
		end = get_ticks() * 1000000000 / f;
		if (!(frame % 10))
			app.fps = 1000000000 / (end - now);
	}
	CloseHandle(timer);
	int_plat_wait_audio();
	prog_deinit();
	deinit_app();
}

void plat_quit(void) {
	app.o = 0;
}

int plat_fps(void) {
	return app.fps;
}

int get_mouse_x() {
	return app.mx;
}

int get_mouse_y() {
	return app.my;
}

int btn_pressed(Btn btn) {
	return held_btns[btn];
}

int btn_just_pressed(Btn btn) {
	return pressed_btns[btn];
}

int btn_just_released(Btn btn) {
	return released_btns[btn];
}

int any_btn_just_pressed(void) {
	int b;
	for (b = 0; b < btn_count; b++)
		if (pressed_btns[b])
			return 1;
	return 0;
}

int mbtn_pressed(MBtn btn) {
	return mheld_btns[btn];
}

int mbtn_just_pressed(MBtn btn) {
	return mpressed_btns[btn];
}

int mbtn_just_released(MBtn btn) {
	return mreleased_btns[btn];
}

int get_mscroll_x() {
	return app.scrx;
}

int get_mscroll_y() {
	return app.scry;
}

const char* get_text_input(int* len) {
	int tl;
	tl = app.txt_len;
	if (!tl) {
		return 0;
	}
	if (len) { *len = tl; }
	app.txt_len = 0;
	return app.txt_buf;
}

int is_running_well() {
	return 1;
}

void print(const char* fmt, ...) {
	va_list args;
	char buf[256];
	va_start(args, fmt);
	vsnprintf(buf, sizeof buf, fmt, args);
	va_end(args);
	MessageBox(
		app.w,
		buf,
		"Info",
		MB_OK | MB_ICONINFORMATION
	);
}

void print_err(const char* fmt, ...) {
	va_list args;
	char buf[256];
	va_start(args, fmt);
	vsnprintf(buf, sizeof buf, fmt, args);
	va_end(args);
	MessageBox(
		app.w,
		buf,
		"Error",
		MB_OK | MB_ICONERROR
	);
}

void print_war(const char* fmt, ...) {
	va_list args;
	char buf[256];
	va_start(args, fmt);
	vsnprintf(buf, sizeof buf, fmt, args);
	va_end(args);
	MessageBox(
		app.w,
		buf,
		"Warning",
		MB_OK | MB_ICONWARNING
	);
}

void pbreak(Error code) {
#if defined(DEBUG) && defined(plat_x86)
	__debugbreak();
	(void)code;
#else
	exit(code);
#endif
}

#endif

#if defined(plat_psvita)

#include <psp2/audioout.h> 
#include <psp2/ctrl.h> 
#include <psp2/display.h>
#include <psp2/io/fcntl.h>
#include <psp2/kernel/processmgr.h> 
#include <psp2/kernel/sysmem.h>
#include <psp2/kernel/threadmgr.h>
#include <psp2/libdbg.h> 
#include <psp2common/kernel/modulemgr.h> 
#include <stdlib.h>
#include <string.h>

const char sceUserMainThreadName[]       = game_name;
const int sceUserMainThreadPriority      = 0x60;
const SceSize sceUserMainThreadStackSize = 0x2000;
const int sceKernelPreloadModuleInhibit  = SCE_KERNEL_PRELOAD_INHIBIT_LIBDBG;

struct File {
	SceUID handle;
	int cursor;
	int size;
};

File* file_open(const char* name, File_Flags flags) {
	File* file;
	SceUID handle;
	int f = 0;
	char buf[256];
	if (flags & file_flags_write)
		f |= SCE_O_WRONLY;
	if (flags & file_flags_read)
		f |= SCE_O_RDONLY;
	if (flags & file_flags_read_write)
		f |= SCE_O_RDWR;
	if (flags & file_flags_create)
		f |= SCE_O_CREAT;
	strcpy(buf, "device:/");
	strcat(buf, name);
	handle = sceIoOpen(name, f, 0777);
	if (handle < 0) {
		return 0;
	}

	file = (File*)malloc(sizeof *file);
	file->handle = handle;
	file->cursor = 0;
	file->size = sceIoLseek(handle, 0, SCE_SEEK_END);
	sceIoLseek(handle, 0, SCE_SEEK_SET);

	return file;
}

void file_close(File* file) {
	sceIoClose(file->handle);
	free(file);
}

void file_read(File* file, void* buffer, int size) {
	sceIoLseek(file->handle, (SceOff)file->cursor, SCE_SEEK_SET);
	file->cursor += (int)sceIoRead(file->handle, buffer, (SceSize)size);
}

void file_write(File* file, const void* buffer, int size) {
	sceIoLseek(file->handle, (SceOff)file->cursor, SCE_SEEK_SET);
	file->cursor += (int)sceIoWrite(file->handle, buffer, (SceSize)size);
}

void file_seek(File* file, int offset, File_Whence whence) {
	switch (whence) {
		case file_whence_begin:
			file->cursor = offset;
			break;
		case file_whence_end:
			file->cursor = file->size - offset;
			break;
		case file_whence_cur:
			file->cursor += offset;
			break;
	}
}

int file_size(File* file) {
	return file->size;
}

int file_exists(const char* name) {
	char buf[256];
	SceUID h;
	strcpy(buf, "device:/");
	strcat(buf, name);
	h = sceIoOpen(name, SCE_O_RDONLY, 0777);
	if (h) {
		sceIoClose(h);
		return 1;
	}
	return 0;
}

void print(const char* fmt, ...) {
	/* todo */
}

void print_err(const char* fmt, ...) {
	/* todo */
}

void print_war(const char* fmt, ...) {
	/* todo */
}

void pbreak(Error code) {
#ifdef DEBUG
	sceDbgAssertionHandler(
		__FILE__,
		__LINE__,
		0,
		SCE_DBG_COMPONENT,
		"pbreak was called: %d",
		code
	);
#else
	sceKernelExitProcess(code);
#endif
}

void panic(void){
  while (1) {
    *(int*)(0xAA) = 0x55; /* trigger coredump */
  }
}

struct {
	int scr_mut;
	int fps, o;
	void* fb;
} app;

static Btn  held_btns[btn_count];
static Btn  pressed_btns[btn_count];
static Btn  released_btns[btn_count];

int btn_pressed(Btn btn) {
	return held_btns[btn];
}

int btn_just_pressed(Btn btn) {
	return pressed_btns[btn];
}

int btn_just_released(Btn btn) {
	return released_btns[btn];
}

int any_btn_just_pressed(void) {
	int b;
	for (b = 0; b < btn_count; b++)
		if (pressed_btns[b])
			return 1;
	return 0;
}

int plat_fps() {
	return app.fps;
}

#define vita_fb_w 960
#define vita_fb_h 544

void init_app(void) {
  SceDisplayFrameBuf fbcfg;
  SceUID block;
	app.o = 0;
	app.scr_mut = sceKernelCreateMutex(
		"screenMutex",
		0,
		0,
		NULL
	);
	block = sceKernelAllocMemBlock(
		"display",
		SCE_KERNEL_MEMBLOCK_TYPE_USER_CDRAM_RW,
		(2 * 1024 * 1024),
		NULL
	);
	sceKernelGetMemBlockBase(block, (void**)&app.fb);
	fbcfg.size = sizeof fbcfg;
	fbcfg.base = app.fb;
	fbcfg.pitch = vita_fb_w;
	fbcfg.pixelformat = SCE_DISPLAY_PIXELFORMAT_A8B8G8R8;
	fbcfg.width = vita_fb_w;
	fbcfg.height = vita_fb_h;
	if (!sceDisplaySetFrameBuf(
		&fbcfg,
		SCE_DISPLAY_SETBUF_NEXTFRAME
	)) {
		pbreak(error_platform_error);
		return;
	}
  app.fps = 0;
	app.o = 1;
}

void deinit_app(void) {

}

void plat_quit(void) {
	app.o = 0;
}

void plat_present(
	int x,
	int y,
	int w,
	int h
) {

}

void plat_flush(void) {
	plat_present(0, 0, renderer_w, renderer_h);
}

int main(void) {
	init_app();
	while (app.o) {

	}
	deinit_app();
}

#endif
