#include "maths.h"
#include "plat.h"
#include "profile.h"
#include "render.h"

static struct {
	Colour pixels[renderer_w * renderer_h];
	Rectangle clip;
} renderer;

Colour make_colour(unsigned rgb, unsigned char a) {
	Colour r;
	r.r = (unsigned char)(rgb >> 16);
	r.g = (unsigned char)(rgb >> 8);
	r.b = (unsigned char)rgb;
	r.a = a;
	return r;
}

Colour make_black() {
	Colour r = { 0, 0, 0, 255 };
	return r;
}

Colour make_white() {
	return make_colour(0xffffff, 255);
}

void init_bitmap(
	Bitmap* bitmap,
	Colour* pixels,
	int w,
	int h
) {
	bitmap->pixels = pixels;
	bitmap->w = w;
	bitmap->h = h;
}

void render_init(void) {
	int i;
	Colour b;
	b = make_black();
	for (i = 0; i < renderer_w * renderer_h; i++)
		renderer.pixels[i] = b;
}

void render_begin(void) {
	int i;
	for (i = 0; i < renderer_w * renderer_h; i++)
		renderer.pixels[i] = make_black();
	render_reset_clip();
}

void render_clear(void) {
	render_clear_col(make_black());
}

void render_clear_col(Colour col) {
	Colour* d;
	int ox, x, y, ex, ey;
	d = renderer.pixels;
	x  = renderer.clip.x;
	y  = renderer.clip.y;
	ex = renderer.clip.w;
	ey = renderer.clip.h;
	ox = x;
	for (; y < ey; y++)
		for (x = ox; x < ex; x++)
			d[x + y * renderer_w] = col;
}

void render_clip(const Rectangle* rect) {
	renderer.clip = *rect;
	renderer.clip.x = maxi(renderer.clip.x, 0);
	renderer.clip.y = maxi(renderer.clip.y, 0);
	renderer.clip.w = mini(renderer.clip.w, renderer_w);
	renderer.clip.h = mini(renderer.clip.h, renderer_h);
}

void render_reset_clip(void) {
	renderer.clip.x = 0;
	renderer.clip.y = 0;
	renderer.clip.w = renderer_w;
	renderer.clip.h = renderer_h;
}

Colour* get_render_pixels(void) {
	return renderer.pixels;
}

Colour blend(Colour dst, Colour src) {
	int ima;

	ima = 0xff - src.a;
	dst.r = (unsigned char)(((src.r * src.a) + (dst.r * ima)) >> 8);
	dst.g = (unsigned char)(((src.g * src.a) + (dst.g * ima)) >> 8);
	dst.b = (unsigned char)(((src.b * src.a) + (dst.b * ima)) >> 8);

	return dst;
}

Colour blend_mod(Colour dst, Colour src, Colour mod) {
	int ima;

	src.a = (src.a * mod.a) >> 8;
	ima = 0xff - src.a;
	dst.r = (unsigned char)(((src.r * mod.r * src.a) >> 16) + ((dst.r * ima) >> 8));
	dst.g = (unsigned char)(((src.g * mod.g * src.a) >> 16) + ((dst.g * ima) >> 8));
	dst.b = (unsigned char)(((src.b * mod.b * src.a) >> 16) + ((dst.b * ima) >> 8));

	return dst;
}

int bm_font_size(BM_Font_Range range) {
	int cnt;
	switch (range) {
		case bm_font_range_printable:
			cnt = 96;
			break;
		case bm_font_range_full:
			cnt = 256;
			break;
	}
	return sizeof(BM_Font) + cnt * sizeof(Rectangle);
}

void init_bm_font(
	BM_Font* font,
	Bitmap* bmp,
	BM_Font_Range range
) {
	int i, x, s, e;
	Colour* px;
	Rectangle r, * rects;
	profiler_begin("init_bm_font");
	x = 0;
	r.y = 1;
	r.h = bmp->h - 1;
	switch (range) {
		case bm_font_range_printable:
			s = ' ';
			e = '~' + 1;
			break;
		case bm_font_range_full:
			s = 0;
			e = 256;
			break;
	}
	font->height = r.h;
	font->first = s;
	font->range = range;
	rects = (Rectangle*)(font + 1);
	for (i = s; i < e; i++) {
		r.x = x;
		r.w = 0;
		for (; x < bmp->w; x++) {
			px = &bmp->pixels[x];
			if (
				px->r == 255 &&
				px->g == 0   &&
				px->b == 0   &&
				px->a == 255
			) {
				px->r = 0;
				px->a = 0;
				r.w++;
				rects[i - s] = r;
				break;
			}
			r.w++;
		}
	}
	font->bmp = bmp;
	profiler_end();
}

void get_font_rect(
	const BM_Font* font,
	Rectangle* dst,
	char c
) {
	const Rectangle* rects;
	rects = (Rectangle*)(font + 1);
	*dst = rects[(unsigned char)c - font->first];
}

Rectangle text_rect(const BM_Font* font, const char* text) {
	Rectangle r = { 0 };
	const Rectangle* rect, * rects;
	const char* c;
	rects = (Rectangle*)(font + 1);
	for (c = text; *c; c++) {
		rect = &rects[(unsigned char)*c - font->first];
		r.h = maxi(r.h, rect->h);
		r.w += rect->w;
	}
	return r;
}

void rfont_char(const BM_Font* font, int x, int y, char c) {
	const Rectangle* rect, * rects;
	rects = (Rectangle*)(font + 1);
	c -= font->first;
	rect = &rects[(unsigned char)c];
	render_bitmap(font->bmp, x, y, rect);
}

void rfont_char_col(
	const BM_Font* font,
	int x,
	int y,
	char c,
	Colour colour
) {
	const Rectangle* rect, * rects;
	rects = (Rectangle*)(font + 1);
	c -= font->first;
	rect = &rects[(unsigned char)c];
	render_bitmap_col(font->bmp, x, y, rect, colour);
}

void rfont_text(
	const BM_Font* font,
	int x,
	int y,
	const char* text
) {
	const char* c;
	char c2;
	const Rectangle* rect, * rects;
	rects = (Rectangle*)(font + 1);
	for (c = text; *c; c++) {
		c2 = *c - font->first;
		rect = &rects[(unsigned char)c2];
		render_bitmap(
			font->bmp,
			x,
			y,
			rect
		);
		x += rect->w;
	}
}

void rfont_text_col(
	const BM_Font* font,
	int x,
	int y,
	const char* text,
	Colour colour
) {
	const char* c;
	char c2;
	const Rectangle* rect, * rects;
	rects = (Rectangle*)(font + 1);
	for (c = text; *c; c++) {
		c2 = *c - font->first;
		rect = &rects[(unsigned char)c2];
		render_bitmap_col(
			font->bmp,
			x,
			y,
			rect,
			colour
		);
		x += rect->w;
	}
}

void init_fixed_font(
	Fixed_Font* font,
	const Bitmap* bmp,
	int char_w,
	int char_h,
	int space
) {
	font->bmp = bmp;
	font->char_w = char_w;
	font->char_h = char_h;
	font->space = space;
}

void rffont_char(
	const Fixed_Font* font,
	int x,
	int y,
	char c
) {
	Rectangle rect;
	c -= ' ';
	rect.x = c * font->char_w;
	rect.y = 0;
	rect.w = font->char_w;
	rect.h = font->char_h;
	render_bitmap(font->bmp, x, y, &rect);
}

void rffont_char_col(
	const Fixed_Font* font,
	int x,
	int y,
	char c,
	Colour colour
) {
	Rectangle rect;
	c -= ' ';
	rect.x = c * font->char_w;
	rect.y = 0;
	rect.w = font->char_w;
	rect.h = font->char_h;
	render_bitmap_col(font->bmp, x, y, &rect, colour);
}

void rffont_text(
	const Fixed_Font* font,
	int x,
	int y,
	const char* text
) {
	const char* c;
	for (c = text; *c; c++) {
		rffont_char(font, x, y, *c);
		x += font->char_w + font->space;
	}
}

void rffont_text_col(
	const Fixed_Font* font,
	int x,
	int y,
	const char* text,
	Colour colour
) {
	const char* c;
	for (c = text; *c; c++) {
		rffont_char_col(font, x, y, *c, colour);
		x += font->char_w + font->space;
	}
}

Rectangle ftext_rect(
	const Fixed_Font* font,
	const char* text
) {
	Rectangle r = { 0 };
	const char* c;
	r.h = font->char_h;
	for (c = text; *c; c++)
		r.w += font->char_w + font->space;
	return r;
}

#define blit_vars(t_) \
	int i, j, stride, sstride, n; \
	Colour* dst; \
	const t_* src; \
	Rectangle sub;
#define blit( \
	src_, \
	dst_, \
	pitch_, \
	clip_, \
	src_w_, \
	src_h_, \
	expr_ \
) \
	profiler_begin("blit"); \
	sub = *rect; \
	if (sub.w <= 0) { goto end; } \
	if (sub.h <= 0) { goto end; } \
	if (sub.w > (src_w_)) { goto end; } \
	if (sub.h > (src_h_)) { goto end; } \
	if ((n = (clip_).x - x) > 0) { \
		sub.w -= n; \
		sub.x += n; \
		x += n; \
	} \
	if ((n = (clip_).y - y) > 0) { \
		sub.h -= n; \
		sub.y += n; \
		y += n; \
	} \
	if ((n = x + sub.w - (clip_).w) > 0) { \
		sub.w -= n; \
	} \
	if ((n = y + sub.h - (clip_).h) > 0) { \
		sub.h -= n; \
	} \
	if (sub.w <= 0) { goto end; } \
	if (sub.h <= 0) { goto end; } \
	dst = (dst_) + (x + y * (pitch_)); \
	src = (src_) + (sub.x + sub.y * (src_w_)); \
	stride = (pitch_) - sub.w; \
	sstride = (src_w_) - sub.w; \
	for (i = 0; i < sub.h; i++) { \
		for (j = 0; j < sub.w; j++) { \
			expr_; \
			dst++; \
			src++; \
		} \
		dst += stride; \
		src += sstride; \
	} \
	end: \
	profiler_end();

#define blit_rect_vars \
	int x1, y1, x2, y2, dr, j, i; \
	Colour* dst;
#define blit_rect(dst_, pitch_, clip_) \
	profiler_begin("blit_rect"); \
	x1 = rect->x < (clip_).x ? (clip_).x : rect->x; \
	y1 = rect->y < (clip_).y ? (clip_).y : rect->y; \
	x2 = rect->x + rect->w; \
	y2 = rect->y + rect->h; \
	x2 = x2 > (clip_).w? (clip_).w: x2; \
	y2 = y2 > (clip_).h? (clip_).h: y2; \
	dst = dst_; \
	dst += x1 + y1 * pitch_; \
	dr = pitch_ - (x2 - x1); \
	for (j = y1; j < y2; j++) { \
		for (i = x1; i < x2; i++) { \
			*dst = blend(*dst, colour); \
			dst++; \
		} \
		dst += dr; \
	} \
	profiler_end();

void render_bitmap(
	const Bitmap* bitmap,
	int x,
	int y,
	const Rectangle* rect
) {
	blit_vars(Colour)
	blit(
		bitmap->pixels,
		renderer.pixels,
		renderer_w,
		renderer.clip,
		bitmap->w,
		bitmap->h,
		*dst = blend(*dst, *src)
	);
}

void render_bitmap_col(
	const Bitmap* bitmap,
	int x,
	int y,
	const Rectangle* rect,
	Colour colour
) {
	blit_vars(Colour)
	blit(
		bitmap->pixels,
		renderer.pixels,
		renderer_w,
		renderer.clip,
		bitmap->w,
		bitmap->h,
		*dst = blend_mod(*dst, *src, colour)
	);
}

void render_mask(
	const unsigned char* pixels,
	int x,
	int y,
	int w,
	int h,
	const Rectangle* rect,
	Colour colour
) {
	blit_vars(unsigned char)
	Colour c;
	c = colour;
	blit(
		pixels,
		renderer.pixels,
		renderer_w,
		renderer.clip,
		w,
		h,
		{
			c.a = colour.a * *src;
			*dst = blend(*dst, c);
		}
	);
}

void render_rect(
	const Rectangle* rect,
	Colour colour
) {
	blit_rect_vars
	blit_rect(
		renderer.pixels,
		renderer_w,
		renderer.clip
	);
}

void rcopy(
	Bitmap* to,
	const Bitmap* from,
	int x,
	int y,
	const Rectangle* rect
) {
	Rectangle clip;
	blit_vars(Colour);
	clip.x = 0;
	clip.y = 0;
	clip.w = to->w;
	clip.h = to->h;
	blit(
		from->pixels,
		to->pixels,
		to->w,
		clip,
		from->w,
		from->h,
		*dst = blend(*dst, *src)
	);
}

void rcopy_col(
	Bitmap* to,
	const Bitmap* from,
	int x,
	int y,
	const Rectangle* rect,
	Colour colour
) {
	Rectangle clip;
	blit_vars(Colour)
	clip.x = 0;
	clip.y = 0;
	clip.w = to->w;
	clip.h = to->h;
	blit(
		from->pixels,
		to->pixels,
		to->w,
		clip,
		from->w,
		from->h,
		*dst = blend_mod(*dst, *src, colour);
	);
}

void rcopy_ac(
	Bitmap* to,
	const Bitmap* from,
	int x,
	int y,
	const Rectangle* rect,
	unsigned char t
) {
	Rectangle clip;
	blit_vars(Colour)
	clip.x = 0;
	clip.y = 0;
	clip.w = to->w;
	clip.h = to->h;
	blit(
		from->pixels,
		to->pixels,
		to->w,
		clip,
		from->w,
		from->h,
		*dst = src->a > t ? *src : *dst;
	);
}

void rcopy_rect(
	Bitmap* to,
	const Rectangle* rect,
	Colour colour
) {
	Rectangle clip;
	blit_rect_vars
	clip.x = 0;
	clip.y = 0;
	clip.w = to->w;
	clip.h = to->h;
	blit_rect(
		to->pixels,
		to->w,
		clip
	);
}
